# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Support for search auto-completion based on postgres'pg_trgm extension."""

from collections import defaultdict

from logilab.common.decorators import monkeypatch
from logilab.database.fti import StopWord, normalize
from logilab.database.postgres import _PGAdvFuncHelper

from cubicweb.server import hook
from cubicweb.server.sources import native


def normalize_words(rawwords):
    words = []
    for word in rawwords:
        try:
            words.append((word, normalize(word)))
        except StopWord:
            continue
    return words


old_sql_init_fti = _PGAdvFuncHelper.sql_init_fti


@monkeypatch(_PGAdvFuncHelper)
def sql_init_fti(self):
    """
    this monkeypatch create `words` table and `pg_trgm` extension
    """
    schema = old_sql_init_fti(self)
    schema += ''';
CREATE EXTENSION  pg_trgm;
CREATE TABLE words (
  etype VARCHAR(64) NOT NULL,
  word VARCHAR(100) NOT NULL,
  normalized_word VARCHAR(100) NOT NULL,
  count INT DEFAULT 1
);

CREATE UNIQUE INDEX words_unique_idx ON words (etype, word);

CREATE INDEX words_word_idx ON words USING gin(word gin_trgm_ops);

ALTER TABLE appears ADD pwords text ARRAY;
'''
    return schema

old_sql_drop_fti = _PGAdvFuncHelper.sql_drop_fti


@monkeypatch(_PGAdvFuncHelper)
def sql_drop_fti(self):
    """this monkeypatch drop `words` table and `pg_trgm` extension
    """
    schema = sql_drop_fti(self)
    schema += ''';
    DROP TABLE words;
    DROP EXTENSION  pg_trgm;
    DROP INDEX words_unique_idx ON words (etype, word);
    '''
    return schema


class HandleWordsIdxOp(hook.DataOperationMixIn, hook.LateOperation):
    containercls = dict
    normalized_form_map = None

    def add_data(self, etype, words, offset):
        """[Un]Register some words for the given offset, depending on the offset (positive for
        addition, negative for deletion).

        `words` may be either a list of string or a dict mapping the original word to its normalized
        form in case of insertion, since the normalized form also have to be recorded.
        """
        etype_words = self._container.setdefault(etype, defaultdict(int))
        for word in words:
            etype_words[word] += offset
        if isinstance(words, dict):
            if self.normalized_form_map is None:
                self.normalized_form_map = {}
            self.normalized_form_map.update(words)

    def precommit_event(self):
        normalized_form_map = self.normalized_form_map
        cursor = self.cnx.cnxset.cu
        args = []
        for etype, words in self.get_data().iteritems():
            for word, offset in words.iteritems():
                if offset != 0:
                    args.append({'etype': etype, 'word': word, 'offset': offset})
                    if offset >= 1:
                        args[-1]['normalized_word'] = normalized_form_map[word]
        if not args:
            return
        cursor.executemany("UPDATE words SET count=count+%(offset)s "
                           "WHERE word=%(word)s AND etype=%(etype)s;",
                           args)
        # remove columns with count < 1
        cursor.execute("DELETE FROM words WHERE count < 1")
        args = (arg for arg in args if arg['offset'] > 0)
        cursor.executemany("INSERT INTO words(word, normalized_word, etype, count) "
                           "SELECT %(word)s, %(normalized_word)s, %(etype)s, %(offset)s "
                           "WHERE NOT EXISTS (SELECT 1 FROM words "
                           "                  WHERE word=%(word)s AND etype=%(etype)s)",
                           args)


orig_fti_unindex_entities = native.NativeSQLSource.fti_unindex_entities


@monkeypatch(native.NativeSQLSource)
def fti_unindex_entities(self, cnx, entities):
    """this monkeypatch register removal of words for synchronization of the`words ` table."""
    if self.dbdriver == 'postgres':
        words_op = HandleWordsIdxOp.get_instance(cnx)
        cursor = cnx.cnxset.cu
        for entity in entities:
            cursor.execute("SELECT pwords FROM appears WHERE uid=%s;" % entity.eid)
            words = cursor.fetchone()
            if words and words[0]:
                words_op.add_data(entity.cw_etype, words[0], -1)
    orig_fti_unindex_entities(self, cnx, entities)


orig_fti_index_entities = native.NativeSQLSource.fti_index_entities


@monkeypatch(native.NativeSQLSource)
def fti_index_entities(self, cnx, entities):
    """this monkeypatch register addition of words for synchronization of the`words ` table."""
    if self.dbdriver == 'postgres':
        words_op = HandleWordsIdxOp.get_instance(cnx)
        cursor = cnx.cnxset.cu
        try:
            for entity in entities:
                words = cursor_index_object(self.dbhelper, entity.eid,
                                            entity.cw_adapt_to('IFTIndexable'), cursor)
                if words:
                    words_op.add_data(entity.cw_etype, words, +1)
        except Exception:  # pylint: disable=broad-except
            self.exception('error while indexing %s', entity)
    else:
        orig_fti_index_entities(self, cnx, entities)


def cursor_index_object(self, uid, obj, cursor):
    """this monkeypatch populates the `words` table"""
    ctx = {'config': self.config, 'uid': int(uid)}
    tsvectors, size, oversized = [], 0, False
    # map original spelling to their normalized form (monkeypatch)
    plain_words = {}
    # sort for test predictability
    for (weight, words) in sorted(obj.get_words().items()):
        normalized_words = normalize_words(words)
        for i, (word, normalized_word) in enumerate(normalized_words):
            size += len(normalized_word) + 1
            if size > self.max_indexed:
                normalized_words = normalized_words[:i]
                oversized = True
                break
            plain_words[word.lower()] = normalized_word
        if normalized_words:
            tsvectors.append("setweight(to_tsvector(%%(config)s, "
                             "%%(wrds_%(w)s)s), '%(w)s')" % {'w': weight})
            ctx['wrds_%s' % weight] = ' '.join(nw for w, nw in normalized_words)
        if oversized:
            break
    if tsvectors:
        ctx['words'] = list(plain_words)
        cursor.execute("INSERT INTO appears(uid, words, weight, pwords) "
                       "VALUES (%%(uid)s, %s, %s, %%(words)s);"
                       % ('||'.join(tsvectors), obj.entity_weight), ctx)
    else:
        plain_words = plain_words.clear()
    return plain_words



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for entities.skos"""

from cubicweb.devtools.testlib import CubicWebTC

import testutils


class ConceptSchemeTC(CubicWebTC):

    def test_optimized_top_concepts(self):
        with self.admin_access.cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(cnx))
            c1 = scheme.add_concept(u'blah')
            c1label = c1.preferred_label[0]
            c2 = c1.add_concept(u'blob')
            cnx.commit()
        # open a fresh connection to ensure all cnx caches are empty
        with self.admin_access.cnx() as cnx:
            scheme = cnx.entity_from_eid(scheme.eid)
            scheme.top_concepts_rset  # simply access to the property
            cache = cnx.transaction_data['ecache']
            self.assertIn(c1.eid, cache)
            self.assertIn(c1label.eid, cache)
            # trick to test the entity's caches: kill its ._cw attribute and
            # attempt to fetch attributes/relations
            c1 = cache[c1.eid]
            del c1._cw
            self.assertEqual([x.eid for x in c1.reverse_broader_concept],
                             [c2.eid])
            self.assertEqual([x.eid for x in c1.preferred_label],
                             [c1label.eid])
            self.assertEqual(set(c1.cw_attr_cache), set(['ark', 'cwuri', 'definition', 'example']))
            c1label = cache[c1label.eid]
            self.assertEqual(set(c1label.cw_attr_cache), set(['kind', 'label', 'language_code']))


if __name__ == '__main__':
    import unittest
    unittest.main()



# coding: utf-8
"""Create concept schemes used by SEDA profile data model"""

from cubes.saem_ref.dataimport.seda import import_seda_schemes


print('-> creating SEDA concept schemes')
import_seda_schemes(cnx)

if repo.system_source.dbdriver == 'postgres':
    # massive store is used, which doesn't handle full-text indexation
    reindex_entities(['ConceptScheme', 'Concept'])



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref workflows definition"""

_ = unicode


def define_publication_workflow(add_workflow, etype):
    """Define a publication workflow for `etype`."""
    wf = add_workflow('publication workflow', etype)
    draft = wf.add_state(_('draft'), initial=True)
    published = wf.add_state(_('published'))
    deprecated = wf.add_state(_('deprecated'))
    wf.add_transition(_('publish'), (draft, ), published,
                      requiredgroups=('managers', 'users'))
    wf.add_transition(_('deprecate'), (published, ), deprecated,
                      requiredgroups=('managers', 'users'))
    return wf



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref postcreate script"""

from cubes.saem_ref import workflows

_ = unicode

set_property('ui.site-title', u'Référentiel SAEM')
set_property('ui.language', u'fr')
set_property('ui.date-format', u'%d/%m/%Y')
set_property('ui.datetime-format', u'%d/%m/%Y %H:%M')

for name in [_('producer'), _('deposit'), _('archival'), _('control'),
             _('enquirer'), _('seda-actor')]:
    create_entity('ArchivalRole', name=name)


for name in [_('person'), _('authority'), _('family'), _('unknown-agent-kind')]:
    create_entity('AgentKind', name=name, cwuri='agentkind/' + name)

create_entity('Organization', name=u'Default authority')

for etype in ('Agent', 'AuthorityRecord', 'ConceptScheme', 'OrganizationUnit',
              'SEDAProfile'):
    wf = workflows.define_publication_workflow(add_workflow, etype)
    if etype == 'SEDAProfile':
        publish = wf.transition_by_name('publish')
        publish.set_permissions(conditions=('U in_group G, G name IN ("users", "managers"),'
                                            'X support_seda_exports ~= "%SEDA-0.2%"',),
                                reset=True)
commit()

if not config.mode == 'test':
    from os.path import join, dirname
    process_script(join(dirname(__file__), 'create_seda_schemes.py'))

for stmt in repo.system_source.dbhelper.sql_create_sequence('ext_ark_count').split(';'):
    if stmt:
        sql(stmt)



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for dataimport"""

import datetime
import os
from os.path import join, dirname, basename
import sys
from itertools import count, imap

from logilab.common.testlib import TestCase

from cubicweb.dataimport.importer import ExtEntity, SimpleImportLog
from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref.dataimport import eac, seda
from cubes.saem_ref.sobjects.skos import _store

import testutils


def mock_(string):
    return string


class EACXMLParserTC(TestCase):

    def test_parse_FRAD033_EAC_00001(self):
        expected = [
            ('AuthorityRecord',
             'FRAD033_EAC_00001',
             {'isni': set([u'22330001300016']),
              'name': set([u'Gironde. Conseil général']),
              'start_date': set([datetime.date(1800, 1, 1)]),
              'end_date': set([datetime.date(2099, 1, 1)]),
              'agent_kind': set(['agentkind/authority']),
             },
            ),
            ('EACSource',
             '1',
             {'source_agent': set(['FRAD033_EAC_00001']),
              'title': set([u'1. Ouvrages imprimés...']),
              'description': set([u'des bouquins']),
              'description_format': set([u'text/plain']),
             },
            ),
            ('EACSource',
             '2',
             {'source_agent': set(['FRAD033_EAC_00001']),
              'url': set([u'http://archives.gironde.fr']),
              'title': set([u'Site des Archives départementales de la Gironde']),
             },
            ),
            ('Activity',
             '3',
             {'type': set([u'create']),
              'generated': set(['FRAD033_EAC_00001']),
              'start': set([datetime.datetime(2013, 4, 24, 5, 34, 41)]),
              'end': set([datetime.datetime(2013, 4, 24, 5, 34, 41)]),
              'description': set([u'bla bla']),
              'description_format': set([u'text/plain']),
             },
            ),
            ('Agent',
             '4',
             {'name': set([u'Delphine Jamet'])},
            ),
            ('Activity',
             '5',
             {'associated_with': set(['4']),
              'generated': set(['FRAD033_EAC_00001']),
              'type': set([u'modify']),
              'start': set([datetime.datetime(2015, 1, 15, 7, 16, 33)]),
              'end': set([datetime.datetime(2015, 1, 15, 7, 16, 33)]),
             },
            ),
            ('AgentKind',
             'agentkind/authority',
             {'name': set([u'authority'])},
            ),
            ('PostalAddress',
             '6',
             {'street': set([u'1 Esplanade Charles de Gaulle']),
              'postalcode': set([u'33074']),
              'city': set([u' Bordeaux Cedex']),
             },
            ),
            ('AgentPlace',
             '7',
             {'name': set([u'Bordeaux (Gironde, France)']),
              'role': set([u'siege']),
              'place_agent': set(['FRAD033_EAC_00001']),
              'place_address': set(['6']),
              'equivalent_concept': set([u'http://catalogue.bnf.fr/ark:/12148/cb152418385']),
             },
            ),
            ('AgentPlace',
             '8',
             {'name': set([u'Toulouse (France)']),
              'place_agent': set(['FRAD033_EAC_00001']),
              'role': set([u'domicile']),
             },
            ),
            ('AgentPlace',
             '9',
             {'name': set([u'Lit']),
              'place_agent': set(['FRAD033_EAC_00001']),
              'role': set([u'dodo']),
             },
            ),
            ('LegalStatus',
             '10',
             {'term': set([u'Collectivité territoriale']),
              'start_date': set([datetime.date(1234, 1, 1)]),
              'end_date': set([datetime.date(3000, 1, 1)]),
              'description': set([u'Description du statut']),
              'description_format': set([u'text/plain']),
              'legal_status_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('Mandate',
             '11',
             {'term': set([u'1. Constitutions françaises']),
              'description': set([u'Description du mandat']),
              'description_format': set([u'text/plain']),
              'mandate_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('History',
             '12',
             {'text': set(["\n".join((
                     u'<p xmlns="urn:isbn:1-931666-33-4" '
                     u'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
                     u'xmlns:xlink="http://www.w3.org/1999/xlink">{0}</p>'
                  ).format(text) for text in [u"La loi du 22 décembre 1789, en divisant ...",
                                              u"L'inspecteur Canardo"])
              ]),
              'text_format': set([u'text/html']),
              'history_agent': set(['FRAD033_EAC_00001']),
              'has_citation': set(['13', '14']),
             },
            ),
            ('Citation',
             '13',
             {'uri': set(['http://www.assemblee-nationale.fr/histoire/images-decentralisation/decentralisation/loi-du-22-decembre-1789-.pdf'])},  # noqa
            ),
            ('Citation',
             '14',
             {'uri': set(['http://pifgadget']), 'note': set(['Voir aussi pifgadget'])},
            ),
            ('Structure',
             '16',
             {'description': set([u'<p xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">Pour accomplir ses missions ...</p>']),  # noqa
              'description_format': set([u'text/html']),
              'structure_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('AgentFunction',
             '17',
             {'description': set([u'<p xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">Quatre grands domaines de compétence...</p>']),  # noqa
              'description_format': set([u'text/html']),
              'function_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('AgentFunction',
             '18',
             {'name': set([u'action sociale']),
              'function_agent': set(['FRAD033_EAC_00001']),
              'description': set([u'<p xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">1. Solidarité\n'  # noqa
                                  u'            blablabla.</p>']),
              'description_format': set([u'text/html']),
              'equivalent_concept': set([
                  u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200'
              ]),
             },
            ),
            ('AgentFunction',
             '19',
             {'name': set([u'environnement']),
              'function_agent': set(['FRAD033_EAC_00001']),
              'equivalent_concept': set([
                  u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074']),
             },
            ),
            ('Occupation',
             '20',
             {'term': set([u'Réunioniste']),
              'start_date': set([datetime.date(1987, 1, 1)]),
              'end_date': set([datetime.date(2099, 1, 1)]),
              'description': set([u'Organisation des réunions ...']),
              'description_format': set([u'text/plain']),
              'occupation_agent': set(['FRAD033_EAC_00001']),
              'has_citation': set(['21']),
              'equivalent_concept': set(['http://pifgadget.com']),
             },
            ),
            ('Citation',
             '21',
             {'note': set([u'la bible']),
             },
            ),
            ('GeneralContext',
             '22',
             {'content': set([u'<p xmlns="urn:isbn:1-931666-33-4" '
                              u'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
                              u'xmlns:xlink="http://www.w3.org/1999/xlink">very famous</p>']),
              'content_format': set([u'text/html']),
              'has_citation': set(['23']),
              'general_context_of': set(['FRAD033_EAC_00001']),
              }
            ),
            ('Citation',
             '23',
             {'note': set([u'it\'s well known']),
             },
            ),
            ('AuthorityRecord',
             'CG33-DIRADSJ',
             {'name': set([u"Gironde. Conseil général. Direction de l'administration et de la "
                           u"sécurité juridique"]),
              'agent_kind': set(['agentkind/unknown-agent-kind']),
             },
            ),
            ('HierarchicalRelation',
             '24',
             {'start_date': set([datetime.date(2008, 1, 1)]),
              'end_date': set([datetime.date(2099, 1, 1)]),
              'description': set([u'Coucou']),
              'description_format': set([u'text/plain']),
              'hierarchical_parent': set(['CG33-DIRADSJ']),
              'hierarchical_child': set(['FRAD033_EAC_00001']),
             },
            ),
            ('AuthorityRecord',
             'whatever',
             {'name': set([u'CG32']),
              'agent_kind': set(['agentkind/unknown-agent-kind']),
             },
            ),
            ('AuthorityRecord',
             '/dev/null',
             {'name': set([u'Trash']),
              'agent_kind': set(['agentkind/unknown-agent-kind']),
             },
            ),
            ('ChronologicalRelation',
             '25',
             {'chronological_predecessor': set(['whatever']),
              'chronological_successor': set(['FRAD033_EAC_00001']),
             },
            ),
            ('ChronologicalRelation',
             '26',
             {'chronological_predecessor': set(['FRAD033_EAC_00001']),
              'chronological_successor': set(['/dev/null']),
              'xml_wrap': set(['<gloups xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">hips</gloups>']),  # noqa
             },
            ),
            ('AssociationRelation',
             '27',
             {'association_from': set(['FRAD033_EAC_00001']),
              'association_to': set(['agent-x']),
             },
            ),
            ('EACResourceRelation',
             '28',
             {'agent_role': set([u'creatorOf']),
              'resource_role': set([u'Fonds d\'archives']),
              'resource_relation_resource': set([
                  'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N']),
              'resource_relation_agent': set(['FRAD033_EAC_00001']),
              'start_date': set([datetime.date(1673, 1, 1)]),
              'end_date': set([datetime.date(1963, 1, 1)]),
              'xml_wrap': set(['<he xmlns="urn:isbn:1-931666-33-4" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">joe</he>']),  # noqa
             },
            ),
            ('ExternalUri',
             'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N',
             {'uri': set([u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N']),
              'cwuri': set([u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N'])},
            ),
            ('ExternalUri',
             'agent-x',
             {'uri': set([u'agent-x']), 'cwuri': set([u'agent-x'])},
            ),
            ('ExternalUri',
             'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200',
             {'uri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200']),
              'cwuri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200'])},
            ),
            ('ExternalUri',
             'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074',
             {'uri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074']),
              'cwuri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074'])},
            ),
            ('ExternalUri',
             'http://catalogue.bnf.fr/ark:/12148/cb152418385',
             {'uri': set([u'http://catalogue.bnf.fr/ark:/12148/cb152418385']),
              'cwuri': set([u'http://catalogue.bnf.fr/ark:/12148/cb152418385'])},
            ),
            ('ExternalUri',
             'http://pifgadget.com',
             {'uri': set([u'http://pifgadget.com']),
              'cwuri': set([u'http://pifgadget.com'])},
            ),
        ]
        expected = [ExtEntity(*vals) for vals in expected]
        fpath = self.datapath('EAC/FRAD033_EAC_00001_simplified.xml')
        import_log = SimpleImportLog(fpath)
        # Use a predictable extid_generator.
        extid_generator = imap(str, count()).next
        importer = eac.EACCPFImporter(fpath, import_log, mock_,
                                      extid_generator=extid_generator)
        entities = list(importer.external_entities())
        self.check_external_entities(entities, expected)
        visited = set([])
        for x in importer._visited.values():
            visited.update(x)
        self.assertItemsEqual(visited, [x.extid for x in expected])
        # Gather not-visited tag by name and group source lines.
        not_visited = {}
        for tagname, sourceline in importer.not_visited():
            not_visited.setdefault(tagname, set([])).add(sourceline)
        self.assertEqual(not_visited,
                         {'maintenanceStatus': set([10]),
                          'publicationStatus': set([12]),
                          'maintenanceAgency': set([14]),
                          'languageDeclaration': set([19]),
                          'conventionDeclaration': set([23, 33, 42]),
                          'localControl': set([52]),
                          'nameEntry': set([112]),  # alternative form.
                          'source': set([74]),  # empty.
                          'biogHist': set([201]),  # empty.
                          'structureOrGenealogy': set([186]),  # empty.
                          })

    def test_mandate_under_mandates(self):
        """In FRAD033_EAC_00003.xml, <mandate> element are within <mandates>."""
        fpath = self.datapath('EAC/FRAD033_EAC_00003.xml')
        import_log = SimpleImportLog(fpath)
        # Use a predictable extid_generator.
        extid_generator = imap(str, count()).next
        importer = eac.EACCPFImporter(fpath, import_log, mock_,
                                      extid_generator=extid_generator)
        entities = list(importer.external_entities())
        expected_terms = [
            u'Code du patrimoine, Livre II',
            u'Loi du 5 brumaire an V [26 octobre 1796]',
            (u'Loi du 3 janvier 1979 sur les archives, accompagnée de ses décrets\n'
             u'                        d’application datant du 3 décembre.'),
            u'Loi sur les archives du 15 juillet 2008',
        ]
        self.assertCountEqual([next(iter(x.values['term'])) for x in entities
                               if x.etype == 'Mandate' and 'term' in x.values],
                              expected_terms)
        mandate_with_link = next(x for x in entities if x.etype == 'Mandate' and
                                 u'Code du patrimoine, Livre II' in x.values['term'])
        extid = next(iter(mandate_with_link.values['has_citation']))
        url = u'http://www.legifrance.gouv.fr/affichCode.do?idArticle=LEGIARTI000019202816'
        citation = next(x for x in entities if x.etype == 'Citation'
                        and url in x.values['uri'])
        self.assertEqual(extid, citation.extid)

    def test_agentfunction_within_functions_tag(self):
        """In FRAD033_EAC_00003.xml, <function> element are within <functions>
        not <description>.
        """
        fpath = self.datapath('EAC/FRAD033_EAC_00003.xml')
        import_log = SimpleImportLog(fpath)
        # Use a predictable extid_generator.
        extid_generator = imap(str, count()).next
        importer = eac.EACCPFImporter(fpath, import_log, mock_,
                                      extid_generator=extid_generator)
        entities = importer.external_entities()
        self.assertCountEqual(
            [x.values['name'].pop() for x in entities
             if x.etype == 'AgentFunction' and 'name' in x.values],
            [u'contr\xf4le', u'collecte', u'classement', u'restauration', u'promotion'])

    def test_no_nameentry_authorizedform(self):
        fpath = self.datapath("EAC/Service de l'administration generale et des assemblees.xml")
        import_log = SimpleImportLog(fpath)
        importer = eac.EACCPFImporter(fpath, import_log, mock_)
        entities = importer.external_entities()
        expected = (u"Gironde. Conseil général. Service de l'administration "
                    u"générale et des assemblées")
        self.assertIn(expected, [x.values['name'].pop() for x in entities
                                 if x.etype == 'AuthorityRecord'])

    def ctx_assert(self, method, actual, expected, ctx, msg=None):
        """Wrap assertion method with a context message"""
        try:
            getattr(self, method)(actual, expected, msg=msg)
        except AssertionError as exc:
            msg = str(exc)
            if ctx:
                msg = ('[%s] ' % ctx) + msg
            raise AssertionError(msg), None, sys.exc_info()[-1]

    def check_external_entities(self, entities, expected):
        entities = extentities2dict(entities)
        expected = extentities2dict(expected)
        etypes, expected_etypes = entities.keys(), expected.keys()
        self.ctx_assert('assertCountEqual', etypes, expected_etypes, ctx='etypes')
        for etype, edict in expected.iteritems():
            entities_etype = entities[etype]
            extids, expected_extids = entities_etype.keys(), edict.keys()
            self.ctx_assert('assertCountEqual', extids, expected_extids,
                            ctx='%s/extids' % etype)
            for extid, values in edict.iteritems():
                self.ctx_assert('assertEqual',
                                tolist(entities_etype[extid]), tolist(values),
                                ctx='%s/%s/values' % (etype, extid))


def tolist(dic):
    """Transform sets in `dic` values as lists for easier comparison."""
    for k, v in dic.iteritems():
        if isinstance(v, set):
            v = list(v)
        dic[k] = v
    return dic


def extentities2dict(entities):
    edict = {}
    for extentity in entities:
        edict.setdefault(extentity.etype, {})[extentity.extid] = extentity.values
    return edict


def eac_import(cnx, fpath):
    import_log = SimpleImportLog(basename(fpath))
    created, updated, _ = cnx.call_service(
        'saem_ref.eac-import', stream=fpath, import_log=import_log,
        raise_on_error=True)
    return created, updated


class EACDataImportTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.user.cw_set(authority=testutils.authority_with_naa(cnx))
            cnx.commit()

    def test_FRAD033_EAC_00001(self):
        fpath = self.datapath('EAC/FRAD033_EAC_00001_simplified.xml')
        with self.admin_access.repo_cnx() as cnx:
            # create a skos concept to ensure it's used instead of a ExternalUri
            scheme = cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(cnx))
            scheme.add_concept(u'environnement',
                               cwuri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074')
            cnx.commit()
            created, updated = eac_import(cnx, fpath)
            self.assertEqual(len(created), 36)
            self.assertEqual(updated, set())
            rset = cnx.find('AuthorityRecord', isni=u'22330001300016')
            self.assertEqual(len(rset), 1)
            agent = rset.one()
            self.assertEqual(agent.kind, 'authority')
            self.assertEqual(agent.start_date, datetime.date(1800, 1, 1))
            self.assertEqual(agent.end_date, datetime.date(2099, 1, 1))
            address = agent.postal_address[0]
            self.assertEqual(address.street, u'1 Esplanade Charles de Gaulle')
            self.assertEqual(address.postalcode, u'33074')
            self.assertEqual(address.city, u' Bordeaux Cedex')
            rset = cnx.execute('Any R,N WHERE P place_agent A, A eid %(eid)s, P role R, P name N',
                               {'eid': agent.eid})
            self.assertCountEqual(rset.rows,
                                  [[u'siege', u'Bordeaux (Gironde, France)'],
                                   [u'domicile', u'Toulouse (France)'],
                                   [u'dodo', u'Lit']])
            self.assertEqual(len(agent.reverse_function_agent), 3)
            for related in ('structure', 'history', 'mandate', 'occupation',
                            'generalcontext', 'legal_status', 'eac_relations',
                            'equivalent_concept', 'control'):
                with self.subTest(related=related):
                    checker = getattr(self, '_check_' + related)
                    checker(cnx, agent)

    def _check_structure(self, cnx, agent):
        rset = cnx.find('Structure', structure_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('description',
                                                    format=u'text/plain').strip(),
                         u'Pour accomplir ses missions ...')

    def _check_history(self, cnx, agent):
        rset = cnx.find('History', history_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('text',
                                                    format=u'text/plain').strip(),
                         u"La loi du 22 décembre 1789, en divisant ...\n\nL'inspecteur Canardo")

    def _check_mandate(self, cnx, agent):
        rset = cnx.find('Mandate', mandate_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('description',
                                                    format=u'text/plain').strip(),
                         u'Description du mandat')

    def _check_occupation(self, cnx, agent):
        occupation = cnx.find('Occupation', occupation_agent=agent).one()
        self.assertEqual(occupation.term, u'Réunioniste')
        citation = occupation.has_citation[0]
        self.assertEqual(citation.note, u'la bible')
        voc = occupation.equivalent_concept[0]
        self.assertEqual(voc.uri, u'http://pifgadget.com')

    def _check_generalcontext(self, cnx, agent):
        occupation = cnx.find('GeneralContext', general_context_of=agent).one()
        self.assertIn(u'very famous', occupation.content)
        self.assertEqual(occupation.content_format, u'text/html')
        citation = occupation.has_citation[0]
        self.assertEqual(citation.note, u'it\'s well known')

    def _check_legal_status(self, cnx, agent):
        rset = cnx.find('LegalStatus', legal_status_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('description',
                                                    format=u'text/plain').strip(),
                         u'Description du statut')

    def _check_eac_relations(self, cnx, agent):
        relation = cnx.find('HierarchicalRelation').one()
        self.assertEqual(relation.printable_value('description',
                                                  format='text/plain'),
                         u'Coucou')
        other_agent = cnx.find('AuthorityRecord', name=u"Gironde. Conseil général. Direction de "
                               u"l'administration et de la sécurité juridique").one()
        self.assertEqual(relation.hierarchical_parent[0], other_agent)
        relation = cnx.find('AssociationRelation').one()
        self.assertEqual(relation.association_from[0], agent)
        other_agent = cnx.find('ExternalUri', uri=u'agent-x').one()
        self.assertEqual(other_agent.cwuri, 'agent-x')
        self.assertEqual(relation.association_to[0], other_agent)
        rset = cnx.find('EACResourceRelation', agent_role=u'creatorOf')
        self.assertEqual(len(rset), 1)
        rrelation = rset.one()
        self.assertEqual(rrelation.resource_relation_agent[0], agent)
        exturi = rrelation.resource_relation_resource[0]
        self.assertEqual(exturi.uri,
                         u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N')
        self.assertEqual(rrelation.xml_wrap.getvalue(),
                         '<he xmlns="urn:isbn:1-931666-33-4" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">joe</he>')  # noqa

    def _check_equivalent_concept(self, cnx, agent):
        functions = dict((f.name, f) for f in agent.reverse_function_agent)
        self.assertEqual(functions['action sociale'].equivalent_concept[0].cwuri,
                         'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200')
        self.assertEqual(functions['action sociale'].equivalent_concept[0].cw_etype,
                         'ExternalUri')
        self.assertEqual(functions['environnement'].equivalent_concept[0].cwuri,
                         'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074')
        self.assertEqual(functions['environnement'].equivalent_concept[0].cw_etype,
                         'Concept')
        self.assertEqual(functions['environnement'].vocabulary_source[0].eid,
                         functions['environnement'].equivalent_concept[0].scheme.eid)
        place = cnx.find('AgentPlace', role=u'siege').one()
        self.assertEqual(place.equivalent_concept[0].cwuri,
                         'http://catalogue.bnf.fr/ark:/12148/cb152418385')

    def _check_control(self, cnx, agent):
        rset = cnx.find('EACSource')
        self.assertEqual(len(rset), 2)
        rset = cnx.execute('Any A WHERE A generated X, X eid %s' % agent.eid)
        # Two activities imported from EAC-CPF file, one created by our hook.
        self.assertEqual(len(rset), 3)
        rset = cnx.execute('Any A WHERE A associated_with X, X name "Delphine Jamet"')
        self.assertEqual(len(rset), 1)

    def test_multiple_imports(self):
        def count_entity(cnx, etype):
            return cnx.execute('Any COUNT(X) WHERE X is %s' % etype)[0][0]

        with self.admin_access.repo_cnx() as cnx:
            nb_records_before = count_entity(cnx, 'AuthorityRecord')
            nb_agents_before = count_entity(cnx, 'Agent')
            for fname in ('FRAD033_EAC_00001.xml', 'FRAD033_EAC_00003.xml',
                          'FRAD033_EAC_00071.xml'):
                fpath = self.datapath('EAC/' + fname)
                created, updated = eac_import(cnx, fpath)
            nb_records_after = count_entity(cnx, 'AuthorityRecord')
            nb_agents_after = count_entity(cnx, 'Agent')
            self.assertEqual(nb_records_after - nb_records_before, 10)
            self.assertEqual(nb_agents_after - nb_agents_before, 5)

    def test_unknown_kind(self):
        with self.admin_access.repo_cnx() as cnx:
            eac_import(cnx, self.datapath('EAC/custom_kind.xml'))
            kind = cnx.find('AgentKind', name=u'a custom kind').one()
            self.assertEqual(kind.cwuri, 'agentkind/a custom kind')
            self.assertEqual(len(kind.reverse_agent_kind), 1)
            # test reimport doesn't attempt to recreate this kind
            eac_import(cnx, self.datapath('EAC/custom_kind2.xml'))


class SEDADataimportTC(CubicWebTC):

    def test_import_seda_scheme(self):
        fpath = join(dirname(__file__), os.pardir, 'migration', 'data', 'seda_dua.csv')
        with self.admin_access.repo_cnx() as cnx:
            store = _store(cnx, naa_what=u'1234')
            seda.create_concept_scheme(cnx, store, u'blabla', fpath,
                                       'seda_appraisal_rule_duration', '', u'ark123')
            rset = cnx.execute('Any C WHERE C is ConceptScheme')
            cs = rset.one()
            self.assertEqual(cs.title, u'blabla')
            self.assertEqual(len(cs.top_concepts), 12)
            concepts = []
            for concept in cs.top_concepts:
                concepts.append((concept.label(), concept.alternative_label[0].label))
            expected_concepts = [(u'P01Y', u'1 an'), (u'P02Y', u'2 ans'), (u'P04Y', u'4 ans'),
                                 (u'P05Y', u'5 ans'), (u'P06Y', u'6 ans'), (u'P10Y', u'10 ans'),
                                 (u'P15Y', u'15 ans'), (u'P20Y', u'20 ans'), (u'P25Y', u'25 ans'),
                                 (u'P30Y', u'30 ans'), (u'P50Y', u'50 ans'), (u'P80Y', u'80 ans')]
            self.assertItemsEqual(concepts, expected_concepts)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-saem-ref automatic tests


uncomment code below if you want to activate automatic test for your cube:

.. sourcecode:: python

    from cubicweb.devtools.testlib import AutomaticWebTest

    class AutomaticWebTest(AutomaticWebTest):
        '''provides `to_test_etypes` and/or `list_startup_views` implementation
        to limit test scope
        '''

        def to_test_etypes(self):
            '''only test views for entities of the returned types'''
            return set(('My', 'Cube', 'Entity', 'Types'))

        def list_startup_views(self):
            '''only test startup views of the returned identifiers'''
            return ('some', 'startup', 'views')
"""

from unittest import TestCase
from logilab.common import attrdict

from cubicweb.devtools import testlib  # noqa

from cubes.saem_ref import cwuri_url


class CWURI_URLTC(TestCase):

    def test(self):
        class entity(attrdict):
            @property
            def _cw(self):
                return self

            def build_url(self, path):
                return 'http://built/' + path

        self.assertEqual(cwuri_url(entity({'cwuri': 'whatever'})),
                         'whatever')
        self.assertEqual(cwuri_url(entity({'cwuri': 'ark:/123'})),
                         'http://built/ark:/123')
        self.assertEqual(cwuri_url(entity({'cwuri': 'http://domain.org/ark:/123'})),
                         'http://domain.org/ark:/123')

if __name__ == '__main__':
    import unittest
    unittest.main()



# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for entities.seda"""

import re
from datetime import date

from lxml import etree

from yams import ValidationError

from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref import cwuri_url

import testutils


def sort_attrs(attrs_string):
    """
    >>> sort_attrs('name="listVersionID" fixed="edition 2009" type="xsd:token" use="required"')
    'fixed="edition 2009" name="listVersionID" type="xsd:token" use="required"'
    """
    return ' '.join(sorted(re.findall('[\w:]+="[^"]*"', attrs_string)))


class SEDAXSDExportTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.publishable_profile(
                cnx, title=u'my profile title &&', description=u'my profile description &&')
            for rtype, label, scheme_title in [
                (u'seda_description_level', u'file',
                 u'SEDA : Niveaux de description de contenu'),
                (u'seda_appraisal_rule_code', u'detruire', None),
                (u'seda_appraisal_rule_duration', u'P10Y', None),
                (u'seda_access_restriction_code', u'AR038', None),
                (u'seda_file_type_code', u'fmt/123', None),
                (u'seda_character_set_code', u'6', None),
                (u'seda_document_type_code', u'CDO', None),
            ]:
                testutils.seda_scheme(cnx, rtype, u'preferred', label,
                                      scheme_title=scheme_title)
            cnx.commit()
            archive = profile.archives[0]
            # Complete the basic content_description created in testutils.
            cd = archive.content_description
            file_concept = testutils.concept(cnx, 'file')
            cd.seda_description_level[0].cw_set(seda_description_level_value=file_concept)
            cd.cw_set(seda_latest_date=cnx.create_entity('SEDADate', value=date(2015, 2, 24)),
                      seda_oldest_date=cnx.create_entity('SEDADate'),
                      seda_description=cnx.create_entity('SEDADescription'))
            cnx.create_entity('SEDAKeyword', seda_keyword_of=cd,
                              seda_keyword_scheme=file_concept.scheme,
                              seda_keyword_value=file_concept)
            cnx.commit()
            # Add appraisal_rule attribute to the archive
            appraisal_code = cnx.create_entity('SEDAAppraisalRuleCode')
            appr_code = testutils.concept(cnx, 'detruire')
            appraisal_code.cw_set(seda_appraisal_rule_code_value=appr_code,
                                  user_annotation=u'detruire le document')
            appr_dur = testutils.concept(cnx, u'P10Y')
            appraisal_duration = cnx.create_entity('SEDAAppraisalRuleDuration',
                                                   seda_appraisal_rule_duration_value=appr_dur,
                                                   user_annotation=u"C'est dans 10ans je m'en irai")
            appraisal = cnx.create_entity('SEDAAppraisalRule',
                                          seda_appraisal_rule_code=appraisal_code,
                                          seda_appraisal_rule_duration=appraisal_duration)
            archive.cw_set(seda_appraisal_rule=appraisal)
            cnx.commit()
            # Add archive object
            ao_appraisal = cnx.create_entity(
                'SEDAAppraisalRule',
                seda_appraisal_rule_code=cnx.create_entity('SEDAAppraisalRuleCode'),
                seda_appraisal_rule_duration=cnx.create_entity('SEDAAppraisalRuleDuration'))
            ao_cd = cnx.create_entity(
                'SEDAContentDescription',
                seda_description_level=cnx.create_entity('SEDADescriptionLevel'))
            ar038 = testutils.concept(cnx, 'AR038')
            ar = cnx.create_entity('SEDAAccessRestrictionCode',
                                   seda_access_restriction_code_value=ar038,
                                   user_annotation=u'restrict')
            cnx.create_entity('ProfileArchiveObject',
                              seda_content_description=ao_cd,
                              seda_appraisal_rule=ao_appraisal,
                              seda_access_restriction_code=ar,
                              seda_name=cnx.create_entity('SEDAName'),
                              seda_parent=archive)
            cnx.commit()
            # Add minimal document and archive object
            file_type_code_value = testutils.concept(cnx, 'fmt/123')
            file_type_code = cnx.create_entity('SEDAFileTypeCode',
                                               user_cardinality=u'1',
                                               seda_file_type_code_value=file_type_code_value)
            character_set_code_value = testutils.concept(cnx, '6')
            character_set_code = cnx.create_entity(
                'SEDACharacterSetCode', user_cardinality=u'1',
                seda_character_set_code_value=character_set_code_value)
            document_type_code_value = testutils.concept(cnx, 'CDO')
            document_type_code = cnx.create_entity(
                'SEDADocumentTypeCode', seda_document_type_code_value=document_type_code_value)
            cnx.create_entity('ProfileDocument',
                              seda_description=cnx.create_entity('SEDADescription'),
                              seda_file_type_code=file_type_code,
                              seda_character_set_code=character_set_code,
                              seda_document_type_code=document_type_code,
                              seda_parent=archive)
            cnx.create_entity('ProfileArchiveObject',
                              seda_name=cnx.create_entity('SEDAName'),
                              seda_parent=archive)
            cnx.commit()
        self.profile_eid = profile.eid

    def test_seda_1_0(self):
        self._test_profile_xsd('SEDA-1.0.xsd', 'test_entities_seda_1.xml')

    def test_seda_0_2(self):
        self._test_profile_xsd('SEDA-0.2.xsd', 'test_entities_seda_02.xml')

    def _test_profile_xsd(self, adapter_id, expected_file):
        def repl(m):
            return '<{0} {1}{2}>'.format(m.group(1), sort_attrs(m.group(2)), m.group(3))

        with self.admin_access.client_cnx() as cnx:
            profile = cnx.entity_from_eid(self.profile_eid)
            adapter = profile.cw_adapt_to(adapter_id)
            generated_xsd = adapter.dump()
            # normalize attributes which are not in the same order (depending on the lxml.etree
            # version)
            generated_xsd = re.sub(r'<([\w:]+) ([^>]+?)(/?)>', repl, generated_xsd)
            with open(self.datapath('xsd', 'XMLSchema.xsd')) as xsd:
                xmlschema = etree.XMLSchema(etree.parse(xsd))
            xml = etree.fromstring(generated_xsd)
            xmlschema.assert_(xml)
            file_concept = testutils.concept(cnx, 'file')
            with open(self.datapath(expected_file)) as expected:
                self.assertMultiLineEqual(expected.read() %
                                          {'concept-uri': cwuri_url(file_concept),
                                           'scheme-ark': file_concept.scheme.ark,
                                           'scheme-uri': cwuri_url(file_concept.scheme)},
                                          generated_xsd)
            # ensure there is no element with @type but a complex type
            namespaces = adapter.namespaces.copy()
            namespaces.pop(None)
            dates = xml.xpath('//xsd:element[@name="Date"]/xsd:complexType/xsd:sequence',
                              namespaces=namespaces)
            self.assertEqual(len(dates), 0)


class SEDATC(CubicWebTC):

    def test_incomplete_profile(self):
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.setup_seda_profile(
                cnx, description=u'my profile description &&')
            cnx.commit()
            # ensure we can remove its first level document unit
            profile.archives[0].cw_delete()
            cnx.commit()
            # now we shouldn't be able to fire the publish transition
            with self.assertRaises(ValidationError):
                profile.cw_adapt_to('IWorkflowable').fire_transition('publish')

    def test_blockers(self):
        with self.admin_access.client_cnx() as cnx:
            profile = cnx.create_entity('SEDAProfile', ark_naa=testutils.naa(cnx))
            blockers = list(profile.cw_adapt_to('SEDA-1.0.xsd').blockers())
            self.assertEqual(blockers,
                             [(profile.eid, 'the profile should have at least a document unit')])
            self.assertFalse(profile.cw_adapt_to('SEDA-1.0.xsd').is_compatible())
            with cnx.deny_all_hooks_but():
                ao = cnx.create_entity('ProfileArchiveObject',
                                       user_cardinality=u'0..1',
                                       seda_name=cnx.create_entity('SEDAName'),
                                       seda_parent=profile)
            profile.cw_clear_all_caches()
            blockers = list(profile.cw_adapt_to('SEDA-1.0.xsd').blockers())
            expected = [
                (ao.eid, '0..1 and 0..n cardinalities are forbidden on first-level document unit'),
                (ao.eid, 'first level document unit must have a content description defined'),
                (ao.eid, 'first level document unit must have an access restriction defined'),
            ]
            self.assertEqual(blockers, expected)
            ao.cw_set(seda_content_description=cnx.create_entity(
                'SEDAContentDescription',
                user_cardinality=u'0..1',
                seda_description_level=cnx.create_entity('SEDADescriptionLevel')),
                seda_access_restriction_code=cnx.create_entity(
                    'SEDAAccessRestrictionCode', user_cardinality=u'0..1'),
                user_cardinality=u'1')
            ao.cw_clear_all_caches()
            blockers = list(profile.cw_adapt_to('SEDA-1.0.xsd').blockers())
            expected = [
                (ao.eid, '0..1 cardinality is not allowed on content description of a first-level document unit'),  # noqa
                (ao.eid, '0..1 cardinality is not allowed on access restriction of a first-level document unit'),  # noqa
            ]
            self.assertEqual(blockers, expected)


class ProfileDocumentTC(CubicWebTC):

    def test_profile_document_dctitle(self):
        with self.admin_access.client_cnx() as cnx:
            testutils.seda_scheme(cnx, u'seda_document_type_code', u'preferred', u'CDO',
                                  scheme_title=None)
            document_type_code_value = testutils.concept(cnx, u'CDO')
            document_type_code = cnx.create_entity(
                'SEDADocumentTypeCode', seda_document_type_code_value=document_type_code_value)
            document = cnx.create_entity('ProfileDocument',
                                         seda_document_type_code=document_type_code)
            self.assertEqual(document.user_annotation, None)
            self.assertEqual(document.dc_title(), 'CDO #%s' % document.eid)
            document.cw_attr_cache['user_annotation'] = 'hi there'
            self.assertEqual(document.dc_title(), 'hi there (CDO)')


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for site_cubicweb"""

from datetime import datetime

import unittest

from cubicweb import devtools  # noqa

from cubes.saem_ref.site_cubicweb import print_tzdatetime_local


class PrintTZDatetimeTC(unittest.TestCase):

    def test(self):
        class req(object):
            @staticmethod
            def property_value(prop):
                return {'ui.timezone': 'Europe/Paris',
                        'ui.datetime-format': '%Y/%m/%d %H:%M'}[prop]
        value = datetime(2015, 5, 26, 14, 12)
        self.assertEqual(print_tzdatetime_local(value, req),
                         '2015/05/26 16:12')

if __name__ == '__main__':
    unittest.main()



# coding: utf-8
import os
from os.path import join, dirname, basename

from cubicweb.dataimport.importer import HTMLImportLog

from cubes.skos.sobjects import lcsv_extentities, store_skos_extentities

from cubes.saem_ref.sobjects.skos import _store


def lcsv_import(cnx, store, fpath, scheme_uri):
    import_log = HTMLImportLog(basename(fpath))
    stream = open(fpath)
    entities = lcsv_extentities(stream, scheme_uri, '\t', 'utf-8')
    store_skos_extentities(cnx, store, entities, import_log,
                           raise_on_error=True, extid_as_cwuri=False)


def create_concept_scheme(cnx, store, title, fpath, rtype, comment, ark):
    scheme = cnx.create_entity('ConceptScheme', title=title, ark=ark)
    inserted = cnx.execute(
        'SET X scheme_relation CR WHERE CR name "%s", X eid %%(x)s' % rtype,
        {'x': scheme.eid})
    assert inserted, comment
    lcsv_import(cnx, store, fpath, scheme.cwuri)


def import_seda_schemes(cnx):
    # 25651 = Archives départementales de la Gironde (ADGIRONDE)
    # XXX ensure that:
    # * NAA for those vocabulary is 25651
    # * generated ark are identical from one instance to another
    store = _store(cnx, naa_what='25651')
    for i, (title, rtype, comment, fname) in enumerate([
        (u"SEDA : Codes de restriction d'accès",
         'seda_access_restriction_code',
         'access control',
         'seda_access_control.csv'),
        (u'SEDA : Sort final',
         'seda_appraisal_rule_code',
         'appraisal rule',
         'seda_appraisal_rule_code.csv'),
        (u'SEDA : Niveaux de description de contenu',
         'seda_description_level',
         'description level',
         'seda_description_level.csv'),
        (u'SEDA : Formats de fichier source',
         'seda_file_type_code',
         'file type code',
         'seda_file_type_code.csv'),
        (u'SEDA : Jeu de caractères de codage',
         'seda_character_set_code',
         'character set code',
         'seda_character_set_code.csv'),
        (u'SEDA : Codes des types de contenu',
         'seda_document_type_code',
         'document type code',
         'seda_document_type_code.csv'),
        (u"SEDA: Durée d'utilité administrative",
         'seda_appraisal_rule_duration',
         'appraisal rule duration',
         'seda_dua.csv'),
    ]):
        if not cnx.find('ConceptScheme', title=title):
            fpath = join(dirname(__file__), os.pardir, 'migration', 'data', fname)
            ark = u'25651/v00000000%s' % i
            create_concept_scheme(cnx, store, title, fpath, rtype, comment, ark=ark)
    store.flush()
    store.commit()
    store.finish()



# coding: utf-8
# pylint: disable=W0622
"""cubicweb-saem-ref application packaging information"""

modname = 'saem_ref'
distname = 'cubicweb-saem-ref'

numversion = (0, 12, 1)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = "Référenciel de Système d'Archivage Électronique Mutualisé"
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ =  {'cubicweb': '>= 3.23.0',
                'cubicweb-squareui': None,
                'cubicweb-addressbook': None,
                'cubicweb-compound': '>= 0.3.0',
                'cubicweb-oaipmh': None,
                'cubicweb-relationwidget': '>= 0.3.3',
                'cubicweb-skos': '>= 0.12.0',
                'cubicweb-prov': None,
                'cubicweb-vtimeline': None,
                'cubicweb-signedrequest': None,
                'python-dateutil': None,
                'pytz': None,
                'psycopg2': None,
                'rdflib': '>= 4.1',
               }
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
    ]

from os import listdir as _listdir
from os.path import join, isdir
from glob import glob

THIS_CUBE_DIR = join('share', 'cubicweb', 'cubes', modname)

def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc')
            and not fname.endswith('~')
            and not isdir(join(dirpath, fname))]

data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
    ]
# check for possible extended cube layout
for dname in ('entities', 'views', 'sobjects', 'hooks', 'schema', 'data',
              'wdoc', 'i18n', 'migration', 'lib', 'dataimport', 'scripts',
              'migration/data'):
    if isdir(dname):
        data_files.append([join(THIS_CUBE_DIR, dname), listdir(dname)])
# Note: here, you'll need to add subdirectories if you want
# them to be included in the debian package




# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref "compound"-related functionalities."""
from warnings import warn

from cubicweb import neg_role
from cubicweb.rset import ResultSet
from cubicweb.predicates import is_instance
from cubicweb.entities.adapters import ITreeAdapter

from cubes.compound import structure_def
from cubes.compound.entities import IContainer, IContained


def authority_record_container_def(schema):
    """Define container for AuthorityRecord"""
    return structure_def(schema, 'AuthorityRecord').items()


def org_container_def(schema):
    return structure_def(schema, 'Organization').items()


def scheme_container_def(schema):
    """Define container for ConceptScheme"""
    skiprtypes = ('broader_concept', 'label_of')
    return structure_def(schema, 'ConceptScheme', skiprtypes=skiprtypes).items()


def concept_container_def(schema):
    """Define container for Concept"""
    return structure_def(schema, 'Concept').items()


def seda_profile_container_def(schema):
    """Define container for SEDAProfile"""
    return structure_def(schema, 'SEDAProfile').items()


class IContainedToITree(ITreeAdapter):
    """Map IContained adapters to ITree, additionaly configured with a list of relations leading to
    contained's children
    """

    children_relations = None  # list of (relation type, role) to get entity's children

    @classmethod
    def build_class(cls, etype, children_relations):
        selector = is_instance(etype)
        return type(etype + 'ITree', (cls,),
                    {'__select__': selector,
                     'children_relations': children_relations})

    @property
    def tree_relation(self):
        parent_relation = self.entity.cw_adapt_to('IContained').parent_relation()
        if parent_relation is None:
            return None
        return parent_relation[0]

    @property
    def child_role(self):
        parent_relation = self.entity.cw_adapt_to('IContained').parent_relation()
        if parent_relation is None:
            return None
        return parent_relation[1]

    def parent(self):
        if self.tree_relation is None:
            return None
        return super(IContainedToITree, self).parent()

    def children_rql(self):
        # XXX may have different shapes
        return ' UNION '.join('(%s)' % self.entity.cw_related_rql(rel, role)
                              for rel, role in self.children_relations)

    def _children(self, entities=True):
        if entities:
            res = []
        else:
            res = ResultSet([], '')
            res.req = self._cw
        for rel, role in self.children_relations:
            res += self.entity.related(rel, role, entities=entities)
        return res

    def different_type_children(self, entities=True):
        """Return children entities of different type as this entity.

        According to the `entities` parameter, return entity objects or the
        equivalent result set.
        """
        res = self._children(entities)
        etype = self.entity.cw_etype
        if entities:
            return [e for e in res if e.cw_etype != etype]
        return res.filtered_rset(lambda x: x.cw_etype != etype, self.entity.cw_col)

    def same_type_children(self, entities=True):
        """Return children entities of the same type as this entity.

        According to the `entities` parameter, return entity objects or the
        equivalent result set.
        """
        res = self._children(entities)
        etype = self.entity.cw_etype
        if entities:
            return [e for e in res if e.cw_etype == etype]
        return res.filtered_rset(lambda x: x.cw_etype == etype, self.entity.cw_col)

    def children(self, entities=True, sametype=False):
        """Return children entities.

        According to the `entities` parameter, return entity objects or the
        equivalent result set.
        """
        if sametype:
            return self.same_type_children(entities)
        else:
            return self._children(entities)


class IContainerToITree(IContainedToITree):
    """Map IContainer adapters to ITree, similarly to :class:`IContainedToITree` but considering
    parent as None
    """

    @classmethod
    def build_class(cls, etype, children_relations):
        selector = is_instance(etype)
        return type(etype + 'ITree', (cls,),
                    {'__select__': selector,
                     'children_relations': children_relations})

    def parent(self):
        return None


def tree_def(schema, etype, skiprtypes=(), skipetypes=()):
    """Return a dictionary {etype: [(relation type, role)]} which are reachable through composite
    relations from the root <etype>. Each key gives the name of an entity type, associated to a list
    of relation/role allowing to access to its children (which is expected to be a contained).
    """
    skiprtypes = frozenset(skiprtypes)
    skipetypes = frozenset(skipetypes)
    contained = {}
    if etype not in schema:
        # Would occur, e.g., during migration.
        warn('%s not found in schema, cannot build a tree' % etype,
             RuntimeWarning)
        return {}
    candidates = set([schema[etype]])
    while candidates:
        eschema = candidates.pop()
        assert eschema not in contained, (eschema.type, contained)
        contained[eschema.type] = set()
        for rschema, teschemas, role in eschema.relation_definitions():
            if rschema.meta or rschema in skiprtypes:
                continue
            for rdef in rschema.rdefs.itervalues():
                if rdef.composite != role:
                    continue
                target = getattr(rdef, neg_role(role))
                if target in skipetypes:
                    continue
                if target not in contained:
                    candidates.add(target)
                contained[eschema.type].add((rschema.type, role))
    return contained


NO_TREE_RTYPES = set(('seda_name', 'seda_description', 'seda_access_restriction_code',
                      'seda_appraisal_rule', 'seda_keyword_of', 'seda_description_level',
                      'seda_document_type_code', 'seda_file_type_code', 'seda_character_set_code',
                      'seda_content_description', 'seda_oldest_date', 'seda_latest_date',
                      'wf_info_for'))

PROFILE_CONTAINER_DEF = {}


def registration_callback(vreg):
    vreg.register(IContainer.build_class('Organization'))
    for etype, parent_relations in org_container_def(vreg.schema):
        IContained.register_class(vreg, etype, parent_relations)
    vreg.register(IContainer.build_class('AuthorityRecord'))
    for etype, parent_relations in authority_record_container_def(vreg.schema):
        IContained.register_class(vreg, etype, parent_relations)
    vreg.register(IContainer.build_class('ConceptScheme'))
    for etype, parent_relations in scheme_container_def(vreg.schema):
        if etype == 'Concept':
            # XXX turn parent_relations to a list to ensure broader_concept is considered first
            parent_relations = list(parent_relations)
            parent_relations.insert(0, ('broader_concept', 'subject'))
        IContained.register_class(vreg, etype, parent_relations)
    vreg.register(IContainer.build_class('Concept'))
    for etype, parent_relations in concept_container_def(vreg.schema):
        IContained.register_class(vreg, etype, parent_relations)
    vreg.register(IContainer.build_class('SEDAProfile'))
    for etype, parent_relations in seda_profile_container_def(vreg.schema):
        PROFILE_CONTAINER_DEF[etype] = parent_relations
        IContained.register_class(vreg, etype, parent_relations)
    # additional ITree interface for seda profile elements
    for etype, children_relations in tree_def(vreg.schema, 'SEDAProfile',
                                              skiprtypes=NO_TREE_RTYPES).items():
        if etype == 'SEDAProfile':
            vreg.register(IContainerToITree.build_class(etype, children_relations))
        else:
            vreg.register(IContainedToITree.build_class(etype, children_relations))



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views for AuthorityRecord"""

from functools import partial

from logilab.common.date import ustrftime
from logilab.mtconverter import xml_escape

from cubicweb import tags
from cubicweb.utils import json_dumps, make_uid, JSString
from cubicweb.uilib import cut, js
from cubicweb.view import EntityView
from cubicweb.predicates import (adaptable, has_related_entities, is_instance, match_kwargs,
                                 score_entity)
from cubicweb.web import component, formfields as ff, formwidgets as fw
from cubicweb.web.views import calendar, tabs, uicfg, actions

from cubes.saem_ref import cwuri_url, user_has_authority
from . import (ImportEntityComponent, RelatedEntitiesListView, SubviewsTabView,
               editlinks, external_link)
from .widgets import JQueryIncompleteDatePicker, ConceptAutoCompleteWidget

_ = unicode

abaa = uicfg.actionbox_appearsin_addmenu
pvs = uicfg.primaryview_section
pvds = uicfg.primaryview_display_ctrl
rdc = uicfg.reledit_ctrl
afs = uicfg.autoform_section
aff = uicfg.autoform_field
affk = uicfg.autoform_field_kwargs


for etype in ('AuthorityRecord',
              'AssociationRelation', 'ChronologicalRelation', 'HierarchicalRelation',
              'Mandate', 'LegalStatus', 'EACResourceRelation'):
    affk.set_field_kwargs(etype, 'start_date',
                          widget=JQueryIncompleteDatePicker(update_min='end_date'))
    affk.set_field_kwargs(etype, 'end_date',
                          widget=JQueryIncompleteDatePicker(update_max='start_date',
                                                            default_end=True))


class EACImportComponent(ImportEntityComponent):
    """Component with a link to import an authority record from an EAC-CPF file."""
    __select__ = (ImportEntityComponent.__select__
                  & is_instance('AuthorityRecord')
                  & user_has_authority())
    _('Import AuthorityRecord')  # generate message used by the import component

    @property
    def import_url(self):
        return self._cw.build_url('view', vid='saem_ref.eac-import')


# Forms
for etype, attr in (
    ('AuthorityRecord', 'isni'),
    ('AuthorityRecord', 'ark'),
    ('AgentPlace', 'name'),
    ('AgentPlace', 'role'),
    ('ConceptScheme', 'ark'),
    ('Concept', 'ark'),
    ('Activity', 'type'),
    ('Mandate', 'term'),
    ('LegalStatus', 'term'),
    ('EACResourceRelation', 'agent_role'),
    ('EACResourceRelation', 'resource_role'),
    ('EACSource', 'title'),
    ('EACSource', 'url'),
):
    affk.set_field_kwargs(etype, attr, widget=fw.TextInput({'size': 80}))
    if attr == 'ark':
        affk.set_field_kwargs(etype, attr, required=False)


# Configure edition of entity types supporting vocabulary_source/equivalent_concept

class EquivalentConceptOrTextField(ff.Field):
    def __init__(self, **kwargs):
        super(EquivalentConceptOrTextField, self).__init__(required=True, **kwargs)
        self.help = _('when linked to a vocabulary, the value is enforced to the label of a '
                      'concept in this vocabulary. Remove the vocabulary source if you want to '
                      'type text freely.')

    def get_widget(self, form):
        return ConceptAutoCompleteWidget(self.name, 'vocabulary_source', optional=True)

    def has_been_modified(self, form):
        return True  # handled in process_posted below

    def process_posted(self, form):
        posted = form._cw.form
        text_val = posted.get(self.input_name(form, 'Label'), '').strip()
        equivalent_eid = posted.get(self.input_name(form), '').strip()
        equivalent_eids = set()
        if equivalent_eid:
            equivalent_eids.add(int(equivalent_eid))
        if not (text_val or equivalent_eid):
            raise ff.ProcessFormError(form._cw.__("required field"))
        entity = form.edited_entity
        if not entity.has_eid() or getattr(entity, self.name) != text_val:
            yield (ff.Field(name=self.name, role='subject', eidparam=True), text_val)
        if (not entity.has_eid() and equivalent_eids) \
           or (entity.has_eid() and
               set(x.eid for x in entity.equivalent_concept) != equivalent_eids):
            subfield = ff.Field(name='equivalent_concept', role='subject', eidparam=True)
            # register the association between the value and the field, because on entity creation,
            # process_posted will be recalled on the newly created field, and if we don't do that it
            # won't find the proper value (this is very nasty)
            form.formvalues[(subfield, form)] = equivalent_eids
            yield (subfield, equivalent_eids)


for etype, attrs in [('AgentFunction', ('name', 'description')),
                     ('AgentPlace', ('name', 'role')),
                     ('Mandate', ('term', 'description')),
                     ('LegalStatus', ('term', 'description')),
                     ('Occupation', ('term', 'description'))]:
    aff.tag_subject_of((etype, attrs[0], '*'), EquivalentConceptOrTextField)
    affk.set_fields_order(etype, ('vocabulary_source',) + attrs)
    if attrs[-1] == 'description':
        affk.tag_attribute((etype, 'description'),
                           {'help': _('let this unspecified to see the definition of the '
                                      'related concept if a vocabulary is specifed')})

afs.tag_object_of(('*', 'vocabulary_source', '*'), 'main', 'hidden')
afs.tag_subject_of(('*', 'vocabulary_source', '*'), 'main', 'attributes')
afs.tag_object_of(('*', 'equivalent_concept', '*'), 'main', 'hidden')
# handled by EquivalentConceptOrTextField
afs.tag_subject_of(('*', 'equivalent_concept', '*'), 'main', 'hidden')
pvs.tag_subject_of(('*', 'equivalent_concept', '*'), 'attributes')
pvs.tag_object_of(('*', 'vocabulary_source', '*'), 'hidden')


# Autocomplete authority record's name

class TextInputCheckSimilar(fw.TextInput):
    """Search for similar authority record names in the database and display them to the user"""

    def __init__(self, *args, **kwargs):
        self.data_initfunc = kwargs.pop('data_initfunc')
        super(TextInputCheckSimilar, self).__init__(*args, **kwargs)

    def _render(self, form, field, render):
        req = form._cw
        req.add_js('cubes.saem_ref.js')
        domid = field.dom_id(form, self.suffix)
        data = self.data_initfunc(form, field)
        category = req._("Similar entities:")
        data = json_dumps([{"label": item, "category": category} for sublist in data
                           for item in sublist])
        req.add_onload(u'cw.jqNode("%s").check_similar_values({source: %s});' % (domid, data))
        return super(TextInputCheckSimilar, self)._render(form, field, render)


def get_record_names(form, field):
    """Return names of all the authority records already existing"""
    return form._cw.execute('Any N WHERE A name N, A is AuthorityRecord').rows

affk.set_field_kwargs('AuthorityRecord', 'name',
                      widget=TextInputCheckSimilar({'size': 80}, data_initfunc=get_record_names))


def unrelated_authorityrecord(rtype, form, field, **kwargs):
    """Choices function returning AuthorityRecord choices.

    Choices values are unrelated through `rtype` to edited entity thus
    filtering out possible ExternalURI targets. It also account for __linkto
    info to build a restricted choice when needed.
    """
    try:
        linkto = form.linked_to[rtype, 'subject'][0]
    except (KeyError, IndexError):
        rset = form.edited_entity.unrelated(rtype, targettype='AuthorityRecord')
        return [(x.dc_title(), unicode(x.eid)) for x in rset.entities()]
    else:
        entity = form._cw.entity_from_eid(linkto)
        return [(entity.dc_title(), unicode(entity.eid))]


# Set fields order for authority records relations (autoform and primary view)
for etype, from_rdef, to_rdef in (
    ('AssociationRelation', 'association_from', 'association_to'),
    ('ChronologicalRelation', 'chronological_predecessor', 'chronological_successor'),
    ('HierarchicalRelation', 'hierarchical_parent', 'hierarchical_child'),
):
    affk.set_fields_order(etype, ('description', 'start_date', 'end_date',
                                  from_rdef, to_rdef))
    for rtype in (from_rdef, to_rdef):
        affk.tag_subject_of((etype, rtype, '*'),
                            {'choices': partial(unrelated_authorityrecord, rtype)})

    for index, attr in enumerate(('description', 'start_date',
                                  'end_date', from_rdef, to_rdef)):
        pvds.tag_attribute((etype, attr), {'order': index})

afs.tag_object_of(('*', 'authority_record', 'AuthorityRecord'), 'main', 'hidden')

# Hide computed relations (See CubicWeb ticket #4903918).
afs.tag_subject_of(('AuthorityRecord', 'postal_address', 'PostalAddress'), 'main', 'hidden')
afs.tag_object_of(('AuthorityRecord', 'postal_address', 'PostalAddress'), 'main', 'hidden')

pvs.tag_subject_of(('AuthorityRecord', 'postal_address', '*'), 'hidden')

# Hide relation to the record for these entity types edition form.
for etype, rtype in (
    ('AgentPlace', 'place_agent'),
    ('AgentFunction', 'function_agent'),
    ('LegalStatus', 'legal_status_agent'),
    ('Mandate', 'mandate_agent'),
    ('History', 'history_agent'),
    ('Structure', 'structure_agent'),
    ('Occupation', 'occupation_agent'),
    ('GeneralContext', 'general_context_of'),
    ('EACSource', 'source_agent'),
    ('EACResourceRelation', 'resource_relation_agent'),
):
    afs.tag_subject_of((etype, rtype, '*'), 'main', 'hidden')


class XMLWrapComponent(component.EntityCtxComponent):
    """CtxComponent to display xml_wrap of entities."""
    __select__ = (
        component.EntityCtxComponent.__select__
        & score_entity(lambda x: getattr(x, 'xml_wrap', None))
    )
    __regid__ = 'saem_ref.xml_wrap'
    context = 'navcontentbottom'
    title = _('xml_wrap')

    def render_body(self, w):
        sourcemt, targetmt = 'text/xml', 'text/html'
        data = self.entity.xml_wrap
        w(self.entity._cw_mtc_transform(data.getvalue(),
                                        sourcemt, targetmt, 'utf-8'))


class AuthorityRecordICalendarableAdapter(calendar.ICalendarableAdapter):
    """ICalendarable adapter for AuthorityRecord entity type."""
    __select__ = calendar.ICalendarableAdapter.__select__ & is_instance('AuthorityRecord')

    @property
    def start(self):
        return self.entity.start_date

    @property
    def stop(self):
        return self.entity.end_date


# AuthorityRecord primary views (tabs)

actions.CopyAction.__select__ &= ~is_instance('AuthorityRecord')

pvds.tag_subject_of(('AuthorityRecord', 'agent_kind', '*'), {'vid': 'text'})


class AuthorityRecordTabbedPrimaryView(tabs.TabbedPrimaryView):
    """Tabbed primary view for AuthorityRecord"""
    __select__ = tabs.TabbedPrimaryView.__select__ & is_instance('AuthorityRecord')
    tabs = [
        _('saem_ref_authorityrecord_general_information'),
        _('saem_ref_authorityrecord_description'),
        _('saem_ref_authorityrecord_properties'),
        _('saem_ref_authorityrecord_relations'),
        _('saem_ref_authorityrecord_lifecycle'),
    ]
    default_tab = 'saem_ref_authorityrecord_general_information'


# Citation
pvds.tag_attribute(('Citation', 'uri'), {'vid': 'urlattr'})
# Generic rules for all entity types having an `has_citation` relationship.
abaa.tag_subject_of(('*', 'has_citation', '*'), False)
afs.tag_subject_of(('*', 'has_citation', '*'), 'main', 'inlined')
pvs.tag_object_of(('*', 'has_citation', 'Citation'), 'hidden')


class CitationLinkView(EntityView):
    __regid__ = 'citation-link'
    __select__ = is_instance('Citation')

    def entity_call(self, entity):
        if entity.uri:
            title = entity.note or entity.uri
            desc = cut(entity.note or u'', 50)
            self.w(u'<a class="truncate" href="%s" title="%s">%s</a>' % (
                xml_escape(entity.uri), xml_escape(desc),
                xml_escape(title)))
        else:
            title = (entity.note or
                     u'{0} #{1}'.format(self._cw._('Citation'), entity.eid))
            self.w(u'<i class="truncate">%s</i>' % title)


# In the main tab.

class AuthorityRecordPrimaryTab(tabs.PrimaryTab):
    """Main tab for authority record, just with a different regid."""
    __regid__ = 'saem_ref_authorityrecord_general_information'


class TextAttributeView(EntityView):
    """View for displaying an entity's text attribute.

    This view should be call with a ``text_attr`` parameter indicating which attribute on the entity
    contains the textual information and will display this information in a ``<div>``.

    Additionally, edit buttons will be shown floating right if user has relevant permissions. And if
    ``icon_info`` parameter is ``True``, user will also see an info button redirecting to the entity
    primary view.
    """

    __regid__ = 'saem.text_attribute'
    __select__ = EntityView.__select__ & match_kwargs('text_attr')

    @editlinks(icon_info=False)
    def entity_call(self, entity, text_attr):
        self._cw.add_js(('jquery.js', 'jquery.expander.js', 'cubes.saem_ref.js'))
        self.w(tags.div(entity.printable_value(text_attr), klass='truncate'))


class AsConceptEntityView(EntityView):
    """View for displaying an entity as a SKOS concept (to which the entity is related).

    This view should be called with ``concept_rtype`` indicating the relation to ``Concept`` entity
    type. This relation should have ``1`` or ``?`` cardinality (that is, the entity should be
    related to at most one concept)

    The view will then display a link to concept URI. The link content will be the entity`s
    ``text_attr`` attribute, or the concept label as a fallback.

    Additionally, edit buttons will be shown floating right if user has relevant permissions.
    """

    __regid__ = 'saem.entity_as_concept'
    __select__ = EntityView.__select__ & match_kwargs('concept_rtype', 'text_attr')

    @editlinks(icon_info=False)
    def entity_call(self, entity, concept_rtype, text_attr, details_attr=None,
                    **kwargs):
        # Get related concept label and uri and associated scheme title and uri
        rset = entity.related(concept_rtype)
        concept = rset.one() if rset else None
        # Compute link content
        link_content = entity.printable_value(text_attr)
        if not link_content and concept:  # empty text_attr => use concept label
            link_content = xml_escape(concept.label())
        if not link_content:  # empty text_attr, no label => empty title
            link_content = u''

        if concept:
            link = external_link(link_content, cwuri_url(concept))
            self.w(u'<strong>{0}</strong>'.format(link))
        else:
            self.w(u'<strong>{0}</strong>'.format(tags.span(link_content, escapecontent=False)))
        if concept and concept.cw_etype == 'Concept':  # Could be ExternalURI
            scheme = concept.in_scheme[0]
            source_content = u'{0}: {1}'.format(
                tags.span(self._cw._('vocabulary_source')),
                external_link(scheme.title, cwuri_url(scheme)))
            self.w(tags.div(source_content, klass='small'))
        # Compute details
        details = entity.printable_value(details_attr) if details_attr else None
        if not details and concept and concept.cw_etype == 'Concept':
            # if not specified but linked to a concept with a definition, show the definition
            details = concept.printable_value('definition')
        if details:
            self._cw.add_js(('jquery.js', 'jquery.expander.js', 'cubes.saem_ref.js'))
            self.w(tags.div(details, klass='help-block truncate'))


# AuthorityRecord life-cycle

class AuthorityRecordLifeCycleTab(SubviewsTabView):
    """Life-cycle tab for authority record."""
    __regid__ = 'saem_ref_authorityrecord_lifecycle'
    __select__ = EntityView.__select__ & is_instance('AuthorityRecord')
    subvids = ('prov.activity-generated',
               'prov.activity-associated-with')


# AuthorityRecord EAC-CPF description tab.

class AuthorityRecordDescriptionTab(SubviewsTabView):
    """Tab view gathering EAC-CPF description information of an AuthorityRecord"""
    __regid__ = 'saem_ref_authorityrecord_description'
    __select__ = EntityView.__select__ & is_instance('AuthorityRecord')
    subvids = (
        'saem_ref.authorityrecord.places',
        'saem_ref.authorityrecord.functions',
        'saem_ref.authorityrecord.legal_status',
        'saem_ref.authorityrecord.mandate',
        'saem_ref.authorityrecord.occupation',
        'saem_ref.authorityrecord.generalcontext',
        'saem_ref.authorityrecord.history',
        'saem_ref.authorityrecord.structure',
    )
    rtype_role_targets = [
        ('place_agent', 'object', None),
        ('function_agent', 'object', None),
        ('mandate_agent', 'object', None),
        ('occupation_agent', 'object', None),
        ('general_context_of', 'object', None),
        ('history_agent', 'object', None),
        ('structure_agent', 'object', None),
    ]


pvs.tag_object_of(('*', 'place_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'place_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'place_agent', 'AuthorityRecord'), False)


class AuthorityRecordPlaceView(RelatedEntitiesListView):
    """View for AuthorityRecordPlace, to be display in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.places'
    rtype = 'place_agent'
    subvid = 'saem.agent_place_as_concept'
    _('creating AgentPlace (AgentPlace place_agent AuthorityRecord %(linkto)s)')


class AuthorityRecordPlaceAsConceptView(EntityView):
    """View for displaying an agent place as a SKOS concept

    If the place is related to a SKOS concept, this view will display a link to the concept URI.

    Additionally, edit buttons will be shown floating right if user has relevant permissions.
    """

    __regid__ = 'saem.agent_place_as_concept'

    def entity_call(self, entity, **kwargs):
        # Output role if any
        role = entity.printable_value('role')
        if role:
            self.w(u'<strong>{0}: </strong>'.format(role))
        # Output place's name
        entity.view('saem.entity_as_concept', w=self.w,
                    concept_rtype='equivalent_concept', text_attr='name',
                    **kwargs)
        # Output address details if they exists
        address = entity.place_address[0] if entity.place_address else None
        if address:
            self._cw.view('incontext', rset=address.as_rset(), w=self.w)


pvs.tag_object_of(('*', 'function_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'function_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'function_agent', 'AuthorityRecord'), False)


class AuthorityRecordFunctionView(RelatedEntitiesListView):
    """View for AgentFunction, to be display in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.functions'
    rtype = 'function_agent'
    subvid = 'saem.entity_as_concept'
    subvid_kwargs = {'concept_rtype': 'equivalent_concept', 'text_attr': 'name',
                     'details_attr': 'description'}
    _('creating AgentFunction (AgentFunction function_agent AuthorityRecord %(linkto)s)')


pvs.tag_object_of(('*', 'mandate_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'mandate_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'mandate_agent', 'AuthorityRecord'), False)


class AuthorityRecordMandateView(RelatedEntitiesListView):
    """View for Mandate, to be displayed in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.mandate'
    rtype = 'mandate_agent'
    subvid = 'saem.entity_as_concept'
    subvid_kwargs = {'concept_rtype': 'equivalent_concept',
                     'text_attr': 'term',
                     'details_attr': 'description'}
    _('creating Mandate (Mandate mandate_agent AuthorityRecord %(linkto)s)')


pvs.tag_object_of(('*', 'occupation_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'occupation_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'occupation_agent', 'AuthorityRecord'), False)


class AuthorityRecordOccupationView(RelatedEntitiesListView):
    """View for Occupation, to be displayed in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.occupation'
    rtype = 'occupation_agent'
    subvid = 'saem.entity_as_concept'
    subvid_kwargs = {'concept_rtype': 'equivalent_concept',
                     'text_attr': 'term',
                     'details_attr': 'description'}
    _('creating Occupation (Occupation occupation_agent AuthorityRecord %(linkto)s)')


pvs.tag_object_of(('*', 'general_context_of', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'general_context_of', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'general_context_of', 'AuthorityRecord'), False)


class GeneralContextView(RelatedEntitiesListView):
    """View for GeneralContext, to be displayed in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.generalcontext'
    rtype = 'general_context_of'
    subvid = 'saem.text_attribute'
    subvid_kwargs = {'text_attr': 'content'}
    _('creating GeneralContext (GeneralContext general_context_of AuthorityRecord %(linkto)s)')


class WithCitationViewMixIn(object):
    """View mixin displaying citation information."""
    __select__ = has_related_entities('has_citation')

    def entity_call(self, entity, *args, **kwargs):
        super(WithCitationViewMixIn, self).entity_call(entity, *args, **kwargs)
        rset = entity.related('has_citation')
        if rset:
            self.w(tags.div(u' '.join([self._cw._('Citation_plural'),
                                       self._cw.view('csv', rset,
                                                     subvid='citation-link')]),
                            klass='small'))


class WithCitationAsConceptEntityView(WithCitationViewMixIn, AsConceptEntityView):
    """Extend entity_as_concept view when entity has related citations."""
    __select__ = AsConceptEntityView.__select__ & WithCitationViewMixIn.__select__


class WithCitationTextAttributeView(WithCitationViewMixIn, TextAttributeView):
    """Extend text_attribute view when entity has related citations."""
    __select__ = TextAttributeView.__select__ & WithCitationViewMixIn.__select__


pvs.tag_object_of(('*', 'legal_status_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'legal_status_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'legal_status_agent', 'AuthorityRecord'), False)


class AuthorityRecordLegalStatusView(RelatedEntitiesListView):
    """View for LegalStatus, to be displayed in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.legal_status'
    rtype = 'legal_status_agent'
    subvid = 'saem.entity_as_concept'
    subvid_kwargs = {'concept_rtype': 'equivalent_concept', 'text_attr': 'term',
                     'details_attr': 'description'}
    _('creating LegalStatus (LegalStatus legal_status_agent AuthorityRecord %(linkto)s)')


pvs.tag_object_of(('*', 'history_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'history_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'history_agent', 'AuthorityRecord'), False)


class AuthorityRecordHistoryView(RelatedEntitiesListView):
    """View for History, to be displayed in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.history'
    rtype = 'history_agent'
    subvid = 'saem.text_attribute'
    subvid_kwargs = {'text_attr': 'text'}
    _('creating History (History history_agent AuthorityRecord %(linkto)s)')


pvs.tag_object_of(('*', 'structure_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'structure_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'structure_agent', 'AuthorityRecord'), False)


class AuthorityRecordStructureView(RelatedEntitiesListView):
    """View for Structure, to be displayed in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.structure'
    rtype = 'structure_agent'
    subvid = 'saem.text_attribute'
    subvid_kwargs = {'text_attr': 'description'}
    _('creating Structure (Structure structure_agent AuthorityRecord %(linkto)s)')


# AuthorityRecord views ###############################

class SaemListItemView(EntityView):
    """List item view working with tabid keyword argument."""
    __regid__ = 'saem.listitem'

    @editlinks(icon_info=False)
    def entity_call(self, entity, **kwargs):
        entity.view('listitem', w=self.w)

# AuthorityRecord EAC-CPF "control" (a.k.a. "properties") tab.


class AuthorityRecordPropertiesTab(SubviewsTabView):
    """Tab view gathering EAC-CPF controle information of an AuthorityRecord"""
    __regid__ = 'saem_ref_authorityrecord_properties'
    __select__ = EntityView.__select__ & is_instance('AuthorityRecord')
    subvids = (
        'saem_ref.authorityrecord.sources',
        'saem_ref.authorityrecord.eac_resource_relation',
    )
    rtype_role_targets = [
        ('source_agent', 'object', None),
        ('resource_relation_agent', 'object', None),
    ]


pvs.tag_object_of(('*', 'source_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'source_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'source_agent', 'AuthorityRecord'), False)

pvds.tag_attribute(('EACSource', 'url'), {'vid': 'urlattr'})


class AuthorityRecordEACSourceView(RelatedEntitiesListView):
    """View for EACSource, to be display in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.sources'
    rtype = 'source_agent'
    subvid = 'saem.eacsource-listitem'
    _('creating EACSource (EACSource source_agent AuthorityRecord %(linkto)s)')


class EACSourceListItemView(EntityView):
    """List item view for EACSource."""

    __regid__ = 'saem.eacsource-listitem'
    __select__ = has_related_entities('source_agent', role='subject')

    @editlinks(icon_info=False)
    def entity_call(self, entity):
        if entity.title:
            self.w(tags.h5(entity.title, klass='list-group-item-heading'))
        if entity.description:
            self.w(u'<p class="list-group-item-text">{0}</p>'.format(
                entity.printable_value('description')))
        if entity.url:
            entity.view('saem.external_link', rtype='url', w=self.w)


pvs.tag_object_of(('*', 'resource_relation_agent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'resource_relation_agent', 'AuthorityRecord'), 'main', 'hidden')
abaa.tag_object_of(('*', 'resource_relation_agent', 'AuthorityRecord'), False)

pvds.tag_subject_of(('EACResourceRelation', 'resource_relation_resource', '*'),
                    {'vid': 'saem.external_link'})


class AuthorityRecordEACResourceRelationView(RelatedEntitiesListView):
    """View for EACResourceRelation, to be displayed in the context of an AuthorityRecord"""
    __regid__ = 'saem_ref.authorityrecord.eac_resource_relation'
    rtype = 'resource_relation_agent'
    subvid = 'saem.eacresource-listitem'
    _('creating EACResourceRelation (EACResourceRelation '
      'resource_relation_agent AuthorityRecord %(linkto)s)')


class EACResourceListItemView(EntityView):
    """List item view for EACResource."""

    __regid__ = 'saem.eacresource-listitem'
    __select__ = has_related_entities('resource_relation_agent', role='subject')

    @editlinks(icon_info=False)
    def entity_call(self, entity, **kwargs):
        resource_role = entity.printable_value('resource_role')
        resource = entity.resource
        title = u' '.join([resource_role, resource.view('saem.external_link')])
        self.w(tags.h5(title, escapecontent=False,
                       klass='list-group-item-heading'))
        if entity.agent_role:
            self.w(u'<p class="list-group-item-text"><em>{0}</em> {1}</p>'.format(
                self._cw._('agent_role'), entity.printable_value('agent_role')))
        if entity.description:
            self.w(u'<p class="list-group-item-text">{0}</p>'.format(
                entity.printable_value('description')))
        if entity.start_date or entity.end_date:
            self.w(u'<p class="list-group-item-text">{0} - {1}</p>'.format(
                entity.printable_value('start_date'),
                entity.printable_value('end_date')))


# AuthorityRecord relations tab.

class AuthorityRecordRelationsTab(SubviewsTabView):
    """Tab view for AuthorityRecord relations."""
    __regid__ = 'saem_ref_authorityrecord_relations'
    __select__ = EntityView.__select__ & is_instance('AuthorityRecord')
    subvids = (
        'saem_ref.authorityrecord.association_relations',
        'saem_ref.authorityrecord.chronological_relation',
        'saem_ref.authorityrecord.hierarchical-links',
    )
    rtype_role_targets = [
        ('chronological_predecessor', 'object', None),
        ('chronological_successor', 'object', None),
        ('hierarchical_child', 'object', None),
        ('hierarchical_parent', 'object', None),
        ('association_from', 'object', None),
    ]


pvs.tag_object_of(('*', 'association_from', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'association_from', 'AuthorityRecord'), 'main', 'hidden')
pvs.tag_object_of(('*', 'association_to', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'association_to', 'AuthorityRecord'), 'main', 'hidden')


class AuthorityRecordAssociationRelationView(EntityView):
    """View for association relations to be displayed in the context of an
    AuthorityRecord on either the `from` or `to` side of the relation.
    """
    __regid__ = 'saem_ref.authorityrecord.association_relations'
    __select__ = EntityView.__select__ & (
        has_related_entities('association_from', role='object') |
        has_related_entities('association_to', role='object'))

    def entity_call(self, entity, **kwargs):
        self.w(tags.h2(self._cw._('associated with')))
        rset = self._cw.execute(
            '(Any X WHERE X association_from A, A eid %(eid)s) '
            'UNION (Any X WHERE X association_to A, A eid %(eid)s)',
            {'eid': entity.eid})
        self._cw.view('list', rset=rset, w=self.w,
                      subvid='saem_ref.associationrelation',
                      main_record=entity,
                      __redirectpath=entity.rest_path())


class AssociationRelationView(EntityView):
    """Extended 'oneline' view for AssociationRelation"""
    __regid__ = 'saem_ref.associationrelation'
    __select__ = (EntityView.__select__ &
                  is_instance('AssociationRelation') & match_kwargs('main_record'))

    _('creating AssociationRelation (AssociationRelation association_from '
      'AuthorityRecord %(linkto)s)')

    @editlinks(icon_info=True)
    def entity_call(self, entity, main_record):
        for other_record in (entity.association_from[0], entity.association_to[0]):
            if other_record != main_record:
                self.w(other_record.view('outofcontext'))
                break
        if entity.start_date or entity.end_date:
            self.w(tags.span(u' ({0}-{1})'.format(entity.printable_value('start_date'),
                                                  entity.printable_value('end_date')),
                             klass='text-muted'))
        if entity.description:
            self.w(tags.div(entity.printable_value('description')))


pvs.tag_object_of(('*', 'chronological_predecessor', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'chronological_predecessor', 'AuthorityRecord'), 'main', 'hidden')
pvs.tag_object_of(('*', 'chronological_successor', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'chronological_successor', 'AuthorityRecord'), 'main', 'hidden')


class ChronologicalRelationView(EntityView):
    """Timeline view with authority records involved in a ChronologicalRelation with this authority
    record entity.
    """
    __regid__ = 'saem_ref.authorityrecord.chronological_relation'
    __select__ = EntityView.__select__ & (
        has_related_entities('chronological_successor', role='object') |
        has_related_entities('chronological_predecessor', role='object'))
    title = _('ChronologicalRelation_plural')
    _("creating ChronologicalRelation (ChronologicalRelation "
      "chronological_predecessor AuthorityRecord %(linkto)s)")
    _("creating ChronologicalRelation (ChronologicalRelation "
      "chronological_successor AuthorityRecord %(linkto)s)")

    def entity_call(self, entity, **kwargs):
        self.w(tags.h2(self._cw._(self.title).lower()))
        rset = self._cw.execute('(Any X WHERE X eid %(eid)s) '
                                'UNION (Any S WHERE RP chronological_predecessor X, '
                                '       RP chronological_successor S, X eid %(eid)s) '
                                'UNION (Any P WHERE RS chronological_successor X, '
                                '       RS chronological_predecessor P, X eid %(eid)s)',
                                {'eid': entity.eid})
        json_url = self._cw.build_url('view', rql=entity.as_rset().printable_rql(),
                                      vid='saem_ref.authorityrecord-timeline-json')
        self._cw.view('vtimeline', rset=rset, w=self.w, custom_settings={'source': json_url})


class AuthorityRecordTimelineJsonView(EntityView):
    """JSON view for agent with chronological relations to be used with vtimeline view."""
    __regid__ = 'saem_ref.authorityrecord-timeline-json'
    __select__ = adaptable('ICalendarable') & is_instance('AuthorityRecord')
    template = False
    content_type = 'application/json'
    binary = True

    headers = {
        'headline': '',
        'type': 'default',
        'text': '',
        'asset': {},
    }

    def entity_call(self, entity):
        data = dict(self.headers)
        data['date'] = [
            self.entity_as_date(entity, tag=self._cw._('subject'),
                                text=entity.dc_description()),
        ]
        if entity.reverse_chronological_successor:
            relation = entity.reverse_chronological_successor[0]
            predecessor = relation.chronological_predecessor[0]
            data['date'].append(self.entity_as_date(
                predecessor,
                text=relation.view('saem_ref.timeline-item',
                                   __redirectpath=entity.rest_path()),
                tag=self._cw._('chronological_predecessor'))
            )
        if entity.reverse_chronological_predecessor:
            relation = entity.reverse_chronological_predecessor[0]
            successor = relation.chronological_successor[0]
            data['date'].append(self.entity_as_date(
                successor,
                text=relation.view('saem_ref.timeline-item',
                                   __redirectpath=entity.rest_path()),
                tag=self._cw._('chronological_successor'))
            )
        self.w(json_dumps({'timeline': data}))

    @staticmethod
    def entity_as_date(entity, **kwargs):
        """Return a dict suitable for insertion within the `date` entry of
        TimelineJS JSON structure.
        """
        date = {}
        calendarable = entity.cw_adapt_to('ICalendarable')
        if calendarable and (calendarable.start or calendarable.stop):
            date.update(
                {'headline': entity.view('incontext'),
                 'text': entity.view('vtimeline-itemdescr')}
            )
            if calendarable.start:
                date['startDate'] = ustrftime(calendarable.start, '%Y,%m,%d')
            if calendarable.stop:
                date['endDate'] = ustrftime(calendarable.stop, '%Y,%m,%d')
        date.update(kwargs)
        return date


class ChronologicalRelationTimelineItemView(EntityView):
    """View for 'text' part of date event in timeline view"""
    __regid__ = 'saem_ref.timeline-item'
    __select__ = is_instance('ChronologicalRelation')

    def entity_call(self, entity, **urlparams):
        self.w(entity.printable_value('description'))
        vreg = self._cw.vreg
        eschema = vreg.schema.eschema(entity.cw_etype)
        if eschema.has_perm(self._cw, 'update'):
            self.w(tags.a(title=self._cw._('edit'), klass='icon-pencil',
                          href=entity.absolute_url('edition', **urlparams)))


def _node(entity, rel=None):
    req = entity._cw
    if rel is not None:
        url = entity.absolute_url()
        descr = rel.printable_value('description', format='text/plain')
        if descr:
            descr_url_desc = descr
            descr = cut(descr, 100)
        else:
            descr = req._('<no description specified>')
            descr_url_desc = req._('view this relation')
        descr_url = rel.absolute_url()
        dates = []
        if rel.start_date:
            dates.append(rel.printable_value('start_date'))
        if rel.end_date:
            dates.append(rel.printable_value('end_date'))
    else:
        url = descr = descr_url = descr_url_desc = dates = u''
    title = entity.dc_title()
    return {
        'id': unicode(entity.eid),
        'title': cut(title, 30),
        'title_url': url,
        'title_url_desc': title,
        'descr': descr,
        'descr_url': descr_url,
        'descr_url_desc': descr_url_desc,
        'dates': u' - '.join(dates),
    }


def _add_html_node_content(properties):
    return (u'<p>%(title)s</p><p>%(descr)s</p><p class="dates">%(dates)s</p>' %
            {'title': tags.a(properties['title'], href=properties['title_url'], klass=u'title'),
             'descr': tags.a(properties['descr'], href=properties['descr_url']),
             'dates': properties['dates']})


class AuthorityRecordGraphView(EntityView):
    __select__ = EntityView.__select__ & (
        has_related_entities('hierarchical_parent', role='object') |
        has_related_entities('hierarchical_child', role='object'))

    __regid__ = 'saem_ref.authorityrecord.hierarchical-links'
    title = _('HierarchicalRelation_plural')

    _('creating HierarchicalRelation (HierarchicalRelation hierarchical_child '
      'AuthorityRecord %(linkto)s)')
    _('creating HierarchicalRelation (HierarchicalRelation hierarchical_parent '
      'AuthorityRecord %(linkto)s)')

    def entity_call(self, entity, **kwargs):
        self.w(tags.h2(self._cw._(self.title).lower()))
        self._cw.add_js(('jquery.js', 'jquery.jOrgChart.js'))
        self._cw.add_css('jquery.jOrgChart.css')
        hiera_list = u'<ul id="hierarchical_relations" style="display:none">'
        tags_closure = u'</ul>'
        rset_parents = self._cw.execute('Any P, PR WHERE PR hierarchical_child X, '
                                        'PR hierarchical_parent P, X eid %s' % entity.eid)
        if rset_parents:
            parent = rset_parents.get_entity(0, 0)
            parent_rel = rset_parents.get_entity(0, 1)
            nprops = _node(parent, parent_rel)
            hiera_list += u'<li>' + _add_html_node_content(nprops) + u'<ul>'
            tags_closure = u'</ul></li>' + tags_closure
        nprops = _node(entity)
        hiera_list += u'<li class="main">' + _add_html_node_content(nprops) + u'<ul>'
        tags_closure = u'</ul></li>' + tags_closure
        for child_rel in entity.reverse_hierarchical_parent:
            for child in child_rel.hierarchical_child:
                nprops = _node(child, child_rel)
                hiera_list += u'<li>' + _add_html_node_content(nprops) + u'</li>'
        hiera_list += tags_closure
        self.w(hiera_list)
        domid = make_uid()
        self._cw.add_onload(js.cw.jqNode('hierarchical_relations').jOrgChart(
            JSString('{chartElement: %s}' % js.cw.jqNode(domid))))
        self.w(u'<div id="%s"></div>' % domid)
        if len(rset_parents) > 1:
            self.w(u'<div class="other-parents icon-attention">')
            self.w(self._cw._('AuthorityRecord %s has several parents:') % entity.name)
            self._cw.view('list', rset_parents, w=self.w)
            self.w(u'</div>')


pvs.tag_object_of(('*', 'hierarchical_parent', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'hierarchical_parent', 'AuthorityRecord'), 'main', 'hidden')

pvs.tag_object_of(('*', 'hierarchical_child', 'AuthorityRecord'), 'hidden')
afs.tag_object_of(('*', 'hierarchical_child', 'AuthorityRecord'), 'main', 'hidden')

# AgentPlace
afs.tag_subject_of(('AgentPlace', 'place_address', 'PostalAddress'),
                   'main', 'inlined')

# AgentFunction
affk.set_field_kwargs('AgentFunction', 'name', widget=fw.TextInput({'size': 80}))



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes for exporting agents to eac"""
from functools import wraps

from lxml import etree

from logilab.common.date import ustrftime
from cubicweb.predicates import is_instance

from cubes.saem_ref import (cwuri_url, eactype_mapping, eacaddress_mapping,
                            eacmaintenancetype_mapping)
from cubes.saem_ref.entities import AbstractXmlAdapter


def add_object_xml_wrap(func):
    """Add an objectXMLWrap sub-element to the element returned by `func`."""
    def wrapper(self, entity, *args, **kwargs):
        elem = func(self, entity, *args, **kwargs)
        if entity.xml_wrap is not None:
            objectXMLWrap = self.element('objectXMLWrap', parent=elem)
            objectXMLWrap.append(etree.parse(entity.xml_wrap).getroot())
        return elem
    return wraps(func)(wrapper)


def add_descriptive_note(func):
    @wraps(func)
    def wrapper(self, entity, *args, **kwargs):
        element = func(self, entity, *args, **kwargs)
        if element is not None and entity.description:
            self.element('descriptiveNote', parent=element).extend(
                self._eac_richstring_paragraph_elements(entity, "description"))
        return element
    return wrapper


def add_citation(func):
    @wraps(func)
    def wrapper(self, entity, *args, **kwargs):
        element = func(self, entity, *args, **kwargs)
        for citation in entity.has_citation:
            attrs = {'xlink:type': 'simple'}
            if citation.uri:
                attrs['xlink:href'] = citation.uri
            self.element('citation', parent=element, attributes=attrs, text=citation.note)
        return element
    return wrapper


class AuthorityRecordEACAdapter(AbstractXmlAdapter):
    __regid__ = 'EAC-CPF'
    __select__ = is_instance('AuthorityRecord')

    namespaces = {
        None: u'urn:isbn:1-931666-33-4',
        u'xsi': u'http://www.w3.org/2001/XMLSchema-instance',
        u'xlink': u'http://www.w3.org/1999/xlink',
    }
    datetime_fmt = '%Y-%m-%dT%H:%M:%S'

    @property
    def file_name(self):
        """Return a file name for the dump."""
        if self.entity.ark:
            name = self.entity.ark.replace("/", "_")
        else:
            name = unicode(self.entity.eid)
        return u'EAC_{0}.xml'.format(name)

    def dump(self):
        """Return an XML string representing the given agent using the EAC-CPF schema."""
        # Keep related activities since they are used multiple times
        self.activities = sorted(self.entity.reverse_generated, key=lambda x: x.start, reverse=True)
        # Root element
        eac_cpf_elt = self.element('eac-cpf', attributes={
            'xsi:schemaLocation': ('urn:isbn:1-931666-33-4 '
                                   'http://eac.staatsbibliothek-berlin.de/schema/cpf.xsd')
        })
        # Top elements: control & cpfDescription
        self.control_element(eac_cpf_elt)
        self.cpfdescription_element(eac_cpf_elt)
        tree = etree.ElementTree(eac_cpf_elt)
        return etree.tostring(tree, xml_declaration=True, encoding='utf-8', pretty_print=True)

    def control_element(self, eac_cpf_elt):
        control_elt = self.element('control', parent=eac_cpf_elt)
        self.element('recordId', parent=control_elt, text=unicode(self.entity.eid))
        self.maintenance_status_element(control_elt)
        self.publication_status_element(control_elt)
        self.maintenance_agency_element(control_elt)
        self.language_declaration_element(control_elt)
        self.maintenance_history_element(control_elt)
        self.sources_element(control_elt)

    def cpfdescription_element(self, eac_cpf_elt):
        cpfdescription_elt = self.element('cpfDescription', parent=eac_cpf_elt)
        self.identity_element(cpfdescription_elt)
        self.description_element(cpfdescription_elt)
        self.relations_element(cpfdescription_elt)

    def maintenance_status_element(self, control_elt):
        if any(activity.type == 'modify' for activity in self.activities):
            status = 'revised'
        else:
            status = 'new'
        self.element('maintenanceStatus', parent=control_elt, text=status)

    def publication_status_element(self, control_elt):
        status = {'draft': 'inProcess', 'published': 'approved'}.get(
            self.entity.cw_adapt_to('IWorkflowable').state, 'inProcess')
        self.element('publicationStatus', parent=control_elt, text=status)

    def maintenance_agency_element(self, control_elt):
        agency_elt = self.element('maintenanceAgency', parent=control_elt)
        self.element('agencyName', parent=agency_elt)

    def language_declaration_element(self, control_elt):
        lang_decl_elt = self.element('languageDeclaration', parent=control_elt)
        self.element('language', parent=lang_decl_elt,
                     attributes={'languageCode': 'fre'}, text=u'français')
        self.element('script', parent=lang_decl_elt, attributes={'scriptCode': 'Latn'},
                     text='latin')

    def maintenance_history_element(self, control_elt):
        history_elt = self.element('maintenanceHistory', parent=control_elt)
        for activity in self.activities:
            self.maintenance_event_element(activity, history_elt)

    def sources_element(self, control_elt):
        sources_elt = self.element('sources')
        for eac_source in self.entity.reverse_source_agent:
            sources_elt.append(self.source_element(eac_source))
        if len(sources_elt):
            control_elt.append(sources_elt)

    def identity_element(self, cpfdescription_elt):
        identity_elt = self.element('identity', parent=cpfdescription_elt)
        self._elt_text_from_attr('entityId', self.entity, 'isni', parent=identity_elt)
        type_mapping = dict((v, k) for k, v in eactype_mapping.items())
        eac_type = type_mapping.get(self.entity.kind)
        self.element('entityType', parent=identity_elt, text=eac_type)
        name_entry = self.element('nameEntry', parent=identity_elt)  # XXX: authorizedForm
        self.element('part', parent=name_entry, text=self.entity.name)

    def description_element(self, cpfdescription_elt):
        agent = self.entity
        description_elt = self.element('description', parent=cpfdescription_elt)
        self.exist_dates_element(description_elt)
        for relation, group_tagname, get_element in [
            ('reverse_place_agent', 'places', self.place_element),
            ('reverse_function_agent', 'functions', self.function_element),
            ('reverse_legal_status_agent', 'legalStatuses', self.legal_status_element),
            ('reverse_occupation_agent', 'occupations', self.occupation_element),
            ('reverse_mandate_agent', 'mandates', self.mandate_element),
        ]:
            relateds = getattr(agent, relation)
            if relateds:
                group_elt = self.element(group_tagname, parent=description_elt)
                for related in relateds:
                    group_elt.append(get_element(related))
        for structure in agent.reverse_structure_agent:
            structure_elt = self.structure_element(structure)
            if len(structure_elt):
                description_elt.append(structure_elt)
        for generalcontext in agent.reverse_general_context_of:
            generalcontext_elt = self.generalcontext_element(generalcontext)
            if len(generalcontext_elt):
                description_elt.append(generalcontext_elt)
        for history in agent.reverse_history_agent:
            bioghist_elt = self.bioghist_element(history)
            if len(bioghist_elt):
                description_elt.append(bioghist_elt)

    def relations_element(self, cpfdescription_elt):
        relations_elt = self.element('relations')
        for rtype, rel_rtype, eac_rtype in [
            ('reverse_hierarchical_child', 'hierarchical_parent', 'hierarchical-parent'),
            ('reverse_hierarchical_parent', 'hierarchical_child', 'hierarchical-child'),
            ('reverse_chronological_successor', 'chronological_predecessor', 'temporal-earlier'),
            ('reverse_chronological_predecessor', 'chronological_successor', 'temporal-later'),
            ('reverse_association_to', 'association_from', 'associative'),
            ('reverse_association_from', 'association_to', 'associative')
        ]:
            for relation in getattr(self.entity, rtype):
                relations_elt.append(
                    self.cpfrelation_element(relation, rel_rtype, eac_rtype))
        for resource_relation in self.entity.reverse_resource_relation_agent:
            relations_elt.append(
                self.resource_relation_element(resource_relation))
        if len(relations_elt):
            cpfdescription_elt.append(relations_elt)

    def maintenance_event_element(self, activity, history_elt):
        event_elt = self.element('maintenanceEvent', parent=history_elt)
        type_mapping = dict((v, k) for k, v in eacmaintenancetype_mapping.items())
        activity_type = type_mapping.get(activity.type, 'created')
        self.element('eventType', parent=event_elt, text=activity_type)
        self.element('eventDateTime', parent=event_elt,
                     attributes={'standardDateTime': ustrftime(activity.start,
                                                               fmt=self.datetime_fmt)},
                     text=ustrftime(activity.start, fmt=self.datetime_fmt))
        ass_agent = activity.associated_with
        if ass_agent:
            self.agent_element(ass_agent[0], event_elt)
        else:  # These tags must be present, even if name is empty
            agent_type_elt = self.element('agentType', text='machine')
            event_elt.append(agent_type_elt)
            event_elt.append(self.element('agent'))
        self._elt_text_from_attr('eventDescription', activity, 'description', parent=event_elt)

    @add_descriptive_note
    @add_object_xml_wrap
    def source_element(self, eac_source):
        url = eac_source.url
        attributes = {'xlink:href': url, 'xlink:type': 'simple'} if url else None
        source_elt = self.element('source', attributes=attributes)
        self.element('sourceEntry', parent=source_elt, text=eac_source.title)
        return source_elt

    def exist_dates_element(self, description_elt):
        date_range = self._eac_date_range_xml_elt(self.entity.start_date, self.entity.end_date)
        if date_range is not None:
            exist_dates = self.element('existDates', parent=description_elt)
            exist_dates.append(date_range)

    @add_citation
    def place_element(self, place):
        place_elt = self.element('place')
        for attr, eac_name in [('role', 'placeRole'), ('name', 'placeEntry')]:
            eac_elt = self._elt_text_from_attr(eac_name, place, attr, parent=place_elt)
            if eac_elt is not None and eac_name == 'placeEntry' and place.equivalent_concept:
                eac_elt.attrib['vocabularySource'] = cwuri_url(place.equivalent_concept[0])
        for address in place.place_address:
            self.address_element(address, place_elt)
        return place_elt

    @add_descriptive_note
    @add_citation
    def function_element(self, function):
        function_elt = self.element('function')
        term_elt = self._elt_text_from_attr('term', function, 'name', parent=function_elt)
        if term_elt is not None and function.equivalent_concept:
            term_elt.attrib['vocabularySource'] = cwuri_url(function.equivalent_concept[0])
        return function_elt

    @add_descriptive_note
    @add_citation
    def legal_status_element(self, legal_status):
        legal_status_elt = self.element('legalStatus')
        self._elt_text_from_attr('term', legal_status, 'term', parent=legal_status_elt)
        return legal_status_elt

    @add_descriptive_note
    @add_citation
    def occupation_element(self, occupation):
        occupation_elt = self.element('occupation')
        term_elt = self._elt_text_from_attr('term', occupation, 'term', parent=occupation_elt)
        if term_elt is not None and occupation.equivalent_concept:
            term_elt.attrib['vocabularySource'] = cwuri_url(occupation.equivalent_concept[0])
        self._eac_date_range_xml_elt(occupation.start_date, occupation.end_date,
                                     parent=occupation_elt)
        return occupation_elt

    @add_descriptive_note
    @add_citation
    def mandate_element(self, mandate):
        mandate_elt = self.element('mandate')
        term_elt = self._elt_text_from_attr('term', mandate, 'term', parent=mandate_elt)
        if term_elt is not None and mandate.equivalent_concept:
            term_elt.attrib['vocabularySource'] = cwuri_url(mandate.equivalent_concept[0])
        return mandate_elt

    def structure_element(self, structure):
        structure_elt = self.element('structureOrGenealogy')
        if structure.description:
            structure_elt.extend(self._eac_richstring_paragraph_elements(structure, "description"))
        return structure_elt

    @add_citation
    def generalcontext_element(self, context):
        context_elt = self.element('generalContext')
        if context.content:
            context_elt.extend(self._eac_richstring_paragraph_elements(context, 'content'))
        return context_elt

    @add_citation
    def bioghist_element(self, history):
        bioghist_elt = self.element('biogHist')
        if history.text:
            bioghist_elt.extend(self._eac_richstring_paragraph_elements(history, "text"))
        return bioghist_elt

    @add_descriptive_note
    @add_object_xml_wrap
    def cpfrelation_element(self, relation, cw_rtype, eac_rtype):
        related = relation.related(cw_rtype).one()  # exactly one target (schema)
        relation_elt = self.element('cpfRelation',
                                    attributes={'cpfRelationType': eac_rtype,
                                                'xlink:href': related.absolute_url(),
                                                'xlink:type': 'simple'})
        if related.cw_etype != 'ExternalUri':
            self.element('relationEntry', parent=relation_elt, text=related.name)
        self._eac_date_range_xml_elt(getattr(relation, 'start_date', None),
                                     getattr(relation, 'end_date', None),
                                     parent=relation_elt)
        return relation_elt

    @add_object_xml_wrap
    def resource_relation_element(self, resource_relation):
        resource = resource_relation.resource_relation_resource[0]
        attrs = {
            'xlink:href': resource.uri,
            'xlink:type': 'simple',
        }
        if resource_relation.resource_role:
            attrs['xlink:role'] = resource_relation.resource_role
        if resource_relation.agent_role:
            attrs['resourceRelationType'] = resource_relation.agent_role
        res_rel_elt = self.element('resourceRelation', attributes=attrs)
        self._eac_date_range_xml_elt(resource_relation.start_date, resource_relation.end_date,
                                     parent=res_rel_elt)
        return res_rel_elt

    def agent_element(self, agent, maintenance_event_elt):
        self.element('agentType', maintenance_event_elt, text='human')
        self.element('agent', maintenance_event_elt, text=agent.name)

    def address_element(self, address, place_elt):
        address_elt = self.element('address', parent=place_elt)
        for eac_name, attr in eacaddress_mapping:
            self._elt_text_from_attr('addressLine', address, attr,
                                     attributes={'localType': eac_name}, parent=address_elt)

    #
    # helper methods for lxml
    #

    def _elt_text_from_attr(self, tag_name, entity, attr_name, parent=None, attributes=None):
        """Return an lxml `Element` whose text is the value of the given attribute on the given
        entity.

        If this element is not empty and if ``parent`` is not ``None``, the element will also be
        inserted in the parent XML element.

        If ``attributes`` is not ``None``, these attributes will be added to the returned element.

        Return ``None`` if element is empty.
        """
        value = getattr(entity, attr_name)
        if value is not None:
            elt = self.element(tag_name, parent=parent, attributes=attributes, text=value)
            return elt

    def _eac_date_range_xml_elt(self, start_date, end_date, parent=None):
        """Return an EAC lxml ``'dateRange'`` ``Element`` with the given boundaries."""
        if not start_date and not end_date:
            return
        date_range = self.element('dateRange', parent=parent)
        for dt, eac_name in [(start_date, 'fromDate'), (end_date, 'toDate')]:
            if not dt:
                continue
            self.element(eac_name, parent=date_range,
                         attributes={'standardDate': dt.isoformat()}, text=dt.isoformat())
        return date_range

    def _eac_richstring_paragraph_elements(self, entity, attr_name):
        fmt = getattr(entity, attr_name + "_format")
        if fmt == "text/plain":
            value = getattr(entity, attr_name)
            if not value:
                return []
            return [self.element('p', text=value)]
        else:
            value = entity.printable_value(attr_name)
            if not value:
                return []
            return list(etree.fromstring(u"<root>{0}</root>".format(value)))



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes for AuthorityRecord."""

from cubicweb.entities import AnyEntity, fetch_config


class AuthorityRecord(AnyEntity):
    __regid__ = 'AuthorityRecord'
    fetch_attrs, cw_fetch_order = fetch_config(('name', 'agent_kind', 'ark'))

    @property
    def kind(self):
        """The kind of agent"""
        return self.agent_kind[0].name

    @property
    def printable_kind(self):
        """The kind of agent, for display"""
        return self.agent_kind[0].printable_value('name')


class AgentKind(AnyEntity):
    __regid__ = 'AgentKind'
    fetch_attrs, cw_fetch_order = fetch_config(('name',))


class ChronologicalRelation(AnyEntity):
    __regid__ = 'ChronologicalRelation'

    def dc_description(self):
        if self.description:
            return self.description


class EACResourceRelation(AnyEntity):
    __regid__ = 'EACResourceRelation'

    @property
    def record(self):
        return self.resource_relation_agent[0]

    @property
    def resource(self):
        return self.resource_relation_resource[0]

    def dc_title(self):
        agent_title = self.agent.dc_title()
        if self.agent_role:
            agent_title += u' (%s)' % self.printable_value('agent_role')
        resource_title = self.resource.dc_title()
        if self.resource_role:
            resource_title += u' (%s)' % self.printable_value('resource_role')
        return (self._cw._('Relation from %(from)s to %(to)s ') %
                {'from': agent_title,
                 'to': resource_title})


class SameAsMixIn(object):
    """Mix-in class for entity types supporting vocabulary_source and
    equivalent_concept relations.
    """

    @property
    def scheme(self):
        return self.vocabulary_source and self.vocabulary_source[0] or None

    @property
    def concept(self):
        return self.equivalent_concept and self.equivalent_concept[0] or None


class AgentPlace(SameAsMixIn, AnyEntity):
    __regid__ = 'AgentPlace'


class AgentFunction(SameAsMixIn, AnyEntity):
    __regid__ = 'AgentFunction'


class Mandate(SameAsMixIn, AnyEntity):
    __regid__ = 'Mandate'


class LegalStatus(SameAsMixIn, AnyEntity):
    __regid__ = 'LegalStatus'


class Occupation(SameAsMixIn, AnyEntity):
    __regid__ = 'Occupation'



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Tests for agent entities"""

import datetime

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.rdfio import RDFLibRDFGraph, RDFRegistry

import testutils


class RDFExportTC(testutils.XmlTestMixin, CubicWebTC):
    """Test case for RDF export"""

    def setUp(self):
        super(RDFExportTC, self).setUp()
        self.reg = RDFRegistry()
        self.reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        self.reg.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
        self.reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
        self.reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
        self.reg.register_prefix('lglb', 'http://www.logilab.org/saem/0#')
        self.reg.register_prefix('org', 'http://www.w3.org/ns/org#')
        self.reg.register_prefix('schema', 'http://schema.org/')
        self.reg.register_prefix('vcard', 'http://www.w3.org/2006/vcard/ns#')

    def _rdf_triples(self, entity, adapter_name='RDFPrimary'):
        graph = RDFLibRDFGraph()
        entity.cw_adapt_to(adapter_name).fill(graph)
        return list(graph.triples())

    def assertItemsIn(self, items, container):
        """Check that elements of `items` are in `container`."""
        for item in items:
            self.assertIn(item, container)

    def assertItemsNotIn(self, items, container):
        """Check that elements of `items` are not in `container`."""
        for item in items:
            self.assertNotIn(item, container)

    def uri(self, s):
        return self.reg.normalize_uri(s)

    def test_organization_unit(self):
        with self.admin_access.client_cnx() as cnx:
            agent = testutils.agent(cnx, u'John Doe')
            ou = testutils.organization_unit(cnx, u'Acme Inc.',
                                             archival_roles=['deposit', 'control'],
                                             contact_point=agent)
            triples = self._rdf_triples(ou)
            self.assertItemsIn([
                (ou.absolute_url(), self.uri('rdf:type'), self.uri('foaf:Organization')),
                (ou.absolute_url(), self.uri('foaf:name'), u'Acme Inc.'),
                (ou.absolute_url(), self.uri('dc:identifier'), ou.ark),
                (ou.absolute_url(), self.uri('vcard:role'), u'deposit'),
                (ou.absolute_url(), self.uri('vcard:role'), u'control'),
                (ou.absolute_url(), self.uri('schema:contactPoint'), agent.absolute_url()),
                (agent.absolute_url(), self.uri('rdf:type'), self.uri('foaf:Person')),
                (agent.absolute_url(), self.uri('foaf:name'), u'John Doe'),
                (agent.absolute_url(), self.uri('dc:identifier'), agent.ark),
            ], triples)

    def test_agent(self):
        with self.admin_access.client_cnx() as cnx:
            agent = testutils.agent(cnx, u'Roger Rabbit')
            triples = self._rdf_triples(agent)
            self.assertItemsIn([
                (agent.absolute_url(), self.uri('rdf:type'), self.uri('foaf:Person')),
                (agent.absolute_url(), self.uri('foaf:name'), u'Roger Rabbit'),
                (agent.absolute_url(), self.uri('dc:identifier'), agent.ark),
            ], triples)

    def test_authority_record_foaf_type(self):
        with self.admin_access.client_cnx() as cnx:
            for kind, foaf_kind in ((u'authority', 'Organization'),
                                    (u'person', 'Person'),
                                    (u'family', 'Group')):
                record = testutils.authority_record(cnx, u'Acme', kind=kind)
                triples = self._rdf_triples(record)
                self.assertItemsIn([
                    (record.absolute_url(), self.uri('rdf:type'), self.uri('foaf:' + foaf_kind)),
                    (record.absolute_url(), self.uri('foaf:name'), u'Acme'),
                    (record.absolute_url(), self.uri('dc:identifier'), record.ark),
                ], triples)

    def test_authority_record_places(self):
        with self.admin_access.client_cnx() as cnx:
            record = testutils.authority_record(cnx, u'Acme Inc. Authority')
            work_address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                             postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', role=u'work', place_agent=record,
                              place_address=work_address)
            home_address = cnx.create_entity('PostalAddress', street=u"Place du Capitole",
                                             postalcode=u'31000', city=u'Toulouse')
            cnx.create_entity('AgentPlace', role=u'home', place_agent=record,
                              place_address=home_address)
            triples = self._rdf_triples(record)
            self.assertItemsIn([
                (record.absolute_url(), self.uri('vcard:hasAddress'), work_address.absolute_url()),
                (work_address.absolute_url(), self.uri('vcard:role'), 'work'),
                (record.absolute_url(), self.uri('vcard:hasAddress'), home_address.absolute_url()),
                (home_address.absolute_url(), self.uri('vcard:role'), 'home'),
            ], triples)

    def test_authority_record_with_chronological_relation(self):
        with self.admin_access.client_cnx() as cnx:
            record1 = testutils.authority_record(cnx, u'Acme Inc. Authority')
            record2 = testutils.authority_record(cnx, u'Acme2 Inc. Authority')
            record3 = testutils.authority_record(cnx, u'Acme3 Inc. Authority')
            cnx.create_entity('ChronologicalRelation', chronological_predecessor=record1,
                              chronological_successor=record2)
            cnx.create_entity('ChronologicalRelation', chronological_predecessor=record2,
                              chronological_successor=record3)
            triples = self._rdf_triples(record2)
            self.assertItemsIn([
                (record2.absolute_url(), self.uri('dcterms:isReplacedBy'), record3.absolute_url()),
                (record2.absolute_url(), self.uri('dcterms:replaces'), record1.absolute_url()),
            ], triples)

    def test_authority_record_with_hierarchical_relation_rdf_export(self):
        with self.admin_access.client_cnx() as cnx:
            record1 = testutils.authority_record(cnx, u'Acme Inc. Authority')
            record2 = testutils.authority_record(cnx, u'Acme Group. Authority')
            cnx.create_entity('HierarchicalRelation',
                              hierarchical_parent=record2,
                              hierarchical_child=record1,
                              start_date=datetime.date(2008, 1, 1),
                              end_date=datetime.date(2099, 1, 1))
            relation_url = record1.reverse_hierarchical_child[0].absolute_url()
            time_interval_url = relation_url + '#timeInterval'
            expected_hierarchical_relations = [
                (relation_url, self.uri('org:organization'), record2.absolute_url()),
                (relation_url, self.uri('org:member'), record1.absolute_url()),
                (relation_url, self.uri('org:role'),
                 'http://www.logilab.org/saem/hierarchical_role'),
                (relation_url, self.uri('org:memberDuring'), time_interval_url),
                (time_interval_url, self.uri('schema:startDate'), datetime.date(2008, 1, 1)),
                (time_interval_url, self.uri('schema:endDate'), datetime.date(2099, 1, 1)),
            ]
            triples = self._rdf_triples(record1)
            self.assertItemsIn(expected_hierarchical_relations, triples)
            triples = self._rdf_triples(record2)
            self.assertItemsIn(expected_hierarchical_relations, triples)

    def test_authority_record_with_associative_relation_rdf_export(self):
        with self.admin_access.client_cnx() as cnx:
            record1 = testutils.authority_record(cnx, u'Acme Inc. Authority')
            record2 = cnx.create_entity('ExternalUri', uri=u'agent2', cwuri=u'agent2')
            cnx.create_entity('AssociationRelation',
                              association_from=record1,
                              association_to=record2)
            relation_url = record1.reverse_association_from[0].absolute_url()
            triples = self._rdf_triples(record1)
            self.assertItemsIn([
                (relation_url, self.uri('org:organization'), record1.absolute_url()),
                (relation_url, self.uri('org:member'), record2.absolute_url()),
                (relation_url, self.uri('org:member'), record1.absolute_url()),
                (relation_url, self.uri('org:organization'), record2.absolute_url()),
                (relation_url, self.uri('org:role'),
                 u'http://www.logilab.org/saem/association_role'),
            ], triples)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref Web Services"""

from cubicweb.view import View
from cubicweb.predicates import match_http_method, match_user_groups
from cubicweb.web.views import json

from cubes.saem_ref import user_has_naa


class AssignArkWebService(json.JsonMixIn, View):
    """JSON view to assign a new Ark for external usage."""

    __regid__ = 'saem_ref.ws.ark'
    __select__ = (match_user_groups('users', 'managers')
                  & user_has_naa()
                  & match_http_method('POST'))

    # XXX could check Accept=application/json
    def call(self):
        self.wdata([{'ark': self._cw.call_service('saem_ref.attribute-ark')}])


class AssignArkWebServiceNoConfiguredUser(AssignArkWebService):
    __select__ = match_user_groups('users', 'managers') & ~user_has_naa()

    def call(self):
        self.wdata([{'error': 'User is not associated to an authority.'}])


class AssignArkWebServiceNonPOST(AssignArkWebService):
    __select__ = match_user_groups('users', 'managers')

    def call(self):
        self.wdata([{'error': 'This service is only accessible using POST.'}])


class AssignArkWebServiceNonAuthenticated(AssignArkWebService):
    __select__ = ~match_user_groups('users', 'managers')

    def call(self):
        self.wdata([{'error': 'This service requires authentication.'}])



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref server objects"""

from itertools import imap

from cubicweb import ValidationError
from cubicweb.predicates import match_user_groups
from cubicweb.server import Service
from cubicweb.dataimport import RQLObjectStore
from cubicweb.dataimport.importer import ExtEntitiesImporter, cwuri2eid

from cubes.skos import to_unicode

from cubes.saem_ref import user_has_naa
from cubes.saem_ref.dataimport import eac


class EACImportService(Service):
    """Service to import an AuthorityRecord from an EAC XML file"""
    __regid__ = 'saem_ref.eac-import'
    __select__ = match_user_groups('managers', 'users') & user_has_naa()

    def call(self, stream, import_log, naa=None, **kwargs):
        authority = self._cw.user.authority[0]
        if naa is None:
            naa = authority.ark_naa[0]
        store = RQLObjectStore(self._cw)
        source = self._cw.repo.system_source
        # only consider the system source for EAC related entity types
        extid2eid = cwuri2eid(self._cw, eac.ETYPES_ORDER_HINT, source_eid=source.eid)
        # though concepts may come from any source
        extid2eid.update(cwuri2eid(self._cw, ('Concept',)))
        importer = ExtEntitiesImporter(self._cw.vreg.schema, store,
                                       import_log=import_log,
                                       extid2eid=extid2eid,
                                       etypes_order_hint=eac.ETYPES_ORDER_HINT,
                                       **kwargs)
        extimporter = eac.EACCPFImporter(stream, import_log, self._cw._)
        extentities = extimporter.external_entities()
        extentities = (ee for ee in extentities
                       if not (ee.etype == 'ExternalUri' and ee.extid in extid2eid))

        def handle_agent_kind(extentities):
            """Create agent kind when necessary and remove them from the entity stream, allowing to
            set cwuri properly without attempt to update.
            """
            for extentity in extentities:
                if extentity.etype == 'AgentKind':
                    if extentity.extid not in extid2eid:
                        name = next(iter(extentity.values['name']))
                        kind = self._cw.create_entity('AgentKind', name=name,
                                                      cwuri=unicode(extentity.extid))
                        extid2eid[extentity.extid] = kind.eid
                else:
                    yield extentity
        extentities = handle_agent_kind(extentities)

        def set_authority_or_naa(extentity):
            """insert function to set parent authority in the ext-entities stream"""
            if extentity.etype == 'AuthorityRecord':
                extentity.values['ark_naa'] = set([naa.cwuri])
            elif extentity.etype == 'Agent':
                extentity.values['authority'] = set([authority.cwuri])
            return extentity

        extentities = imap(set_authority_or_naa, extentities)
        extid2eid[naa.cwuri] = naa.eid
        extid2eid[authority.cwuri] = authority.eid

        try:
            importer.import_entities(extentities)
        except Exception as exc:
            self._cw.rollback()
            import_log.record_fatal(to_unicode(exc))
            raise
        else:
            try:
                self._cw.commit()
            except ValidationError as exc:
                import_log.record_error('validation error: ' + to_unicode(exc))
                raise
            else:
                import_log.record_info('%s entities created' % len(importer.created))
                import_log.record_info('%s entities updated' % len(importer.updated))
        if extimporter.record is not None:
            extid = extimporter.record.extid
            record_eid = importer.extid2eid[extid]
        else:
            record_eid = None
        msg = self._cw._('element {0}, line {1} not parsed')
        for tagname, sourceline in extimporter.not_visited():
            import_log.record_warning(msg.format(tagname, sourceline))
        return importer.created, importer.updated, record_eid


class AllocateArk(Service):
    """Service to allocate an ark identifier"""
    __regid__ = 'saem_ref.attribute-ark'
    __select__ = match_user_groups('managers', 'users') & user_has_naa()

    def call(self):
        naa = self._cw.user.naa
        generator = self._cw.vreg['adapters'].select('IARKGenerator', self._cw,
                                                     naa_what=naa.what)
        return generator.generate_ark()
